import unittest

from cubing_algs.vcube import VCube


class TestExplicitMovePermutations(unittest.TestCase):
    @staticmethod
    def create_numbered_cube() -> VCube:
        state = ''.join([chr(ord('A') + i) for i in range(54)])
        return VCube(initial=state, check=False)

    @staticmethod
    def get_permutations(initial_state: str,
                         final_state: str) -> dict[int, int]:
        permutations = {}
        for dest_pos in range(54):
            dest_char = final_state[dest_pos]
            src_pos = initial_state.index(dest_char)
            if dest_pos != src_pos:
                permutations[dest_pos] = src_pos
        return permutations

    def test_U_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move U."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('U')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 6,
            1: 3,
            2: 0,
            3: 7,
            5: 1,
            6: 8,
            7: 5,
            8: 2,
            9: 45,
            10: 46,
            11: 47,
            18: 9,
            19: 10,
            20: 11,
            36: 18,
            37: 19,
            38: 20,
            45: 36,
            46: 37,
            47: 38,
        }

        self.assertEqual(permutations, expected_perms)

    def test_U_prime_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move U'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("U'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 2,
            1: 5,
            2: 8,
            3: 1,
            5: 7,
            6: 0,
            7: 3,
            8: 6,
            9: 18,
            10: 19,
            11: 20,
            18: 36,
            19: 37,
            20: 38,
            36: 45,
            37: 46,
            38: 47,
            45: 9,
            46: 10,
            47: 11,
        }

        self.assertEqual(permutations, expected_perms)

    def test_U_2_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move U2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('U2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 8,
            1: 7,
            2: 6,
            3: 5,
            5: 3,
            6: 2,
            7: 1,
            8: 0,
            9: 36,
            10: 37,
            11: 38,
            18: 45,
            19: 46,
            20: 47,
            36: 9,
            37: 10,
            38: 11,
            45: 18,
            46: 19,
            47: 20,
        }

        self.assertEqual(permutations, expected_perms)

    def test_R_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move R."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('R')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            2: 20,
            5: 23,
            8: 26,
            9: 15,
            10: 12,
            11: 9,
            12: 16,
            14: 10,
            15: 17,
            16: 14,
            17: 11,
            20: 29,
            23: 32,
            26: 35,
            29: 51,
            32: 48,
            35: 45,
            45: 8,
            48: 5,
            51: 2,
        }

        self.assertEqual(permutations, expected_perms)

    def test_R_prime_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move R'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("R'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            2: 51,
            5: 48,
            8: 45,
            9: 11,
            10: 14,
            11: 17,
            12: 10,
            14: 16,
            15: 9,
            16: 12,
            17: 15,
            20: 2,
            23: 5,
            26: 8,
            29: 20,
            32: 23,
            35: 26,
            45: 35,
            48: 32,
            51: 29,
        }

        self.assertEqual(permutations, expected_perms)

    def test_R_2_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move R2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('R2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            2: 29,
            5: 32,
            8: 35,
            9: 17,
            10: 16,
            11: 15,
            12: 14,
            14: 12,
            15: 11,
            16: 10,
            17: 9,
            20: 51,
            23: 48,
            26: 45,
            29: 2,
            32: 5,
            35: 8,
            45: 26,
            48: 23,
            51: 20,
        }

        self.assertEqual(permutations, expected_perms)

    def test_F_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move F."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('F')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            6: 44,
            7: 41,
            8: 38,
            9: 6,
            12: 7,
            15: 8,
            18: 24,
            19: 21,
            20: 18,
            21: 25,
            23: 19,
            24: 26,
            25: 23,
            26: 20,
            27: 15,
            28: 12,
            29: 9,
            38: 27,
            41: 28,
            44: 29,
        }

        self.assertEqual(permutations, expected_perms)

    def test_F_prime_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move F'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("F'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            6: 9,
            7: 12,
            8: 15,
            9: 29,
            12: 28,
            15: 27,
            18: 20,
            19: 23,
            20: 26,
            21: 19,
            23: 25,
            24: 18,
            25: 21,
            26: 24,
            27: 38,
            28: 41,
            29: 44,
            38: 8,
            41: 7,
            44: 6,
        }

        self.assertEqual(permutations, expected_perms)

    def test_F_2_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move F2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('F2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            6: 29,
            7: 28,
            8: 27,
            9: 44,
            12: 41,
            15: 38,
            18: 26,
            19: 25,
            20: 24,
            21: 23,
            23: 21,
            24: 20,
            25: 19,
            26: 18,
            27: 8,
            28: 7,
            29: 6,
            38: 15,
            41: 12,
            44: 9,
        }

        self.assertEqual(permutations, expected_perms)

    def test_D_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move D."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('D')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            15: 24,
            16: 25,
            17: 26,
            24: 42,
            25: 43,
            26: 44,
            27: 33,
            28: 30,
            29: 27,
            30: 34,
            32: 28,
            33: 35,
            34: 32,
            35: 29,
            42: 51,
            43: 52,
            44: 53,
            51: 15,
            52: 16,
            53: 17,
        }

        self.assertEqual(permutations, expected_perms)

    def test_D_prime_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move D'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("D'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            15: 51,
            16: 52,
            17: 53,
            24: 15,
            25: 16,
            26: 17,
            27: 29,
            28: 32,
            29: 35,
            30: 28,
            32: 34,
            33: 27,
            34: 30,
            35: 33,
            42: 24,
            43: 25,
            44: 26,
            51: 42,
            52: 43,
            53: 44,
        }

        self.assertEqual(permutations, expected_perms)

    def test_D_2_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move D2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('D2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            15: 42,
            16: 43,
            17: 44,
            24: 51,
            25: 52,
            26: 53,
            27: 35,
            28: 34,
            29: 33,
            30: 32,
            32: 30,
            33: 29,
            34: 28,
            35: 27,
            42: 15,
            43: 16,
            44: 17,
            51: 24,
            52: 25,
            53: 26,
        }

        self.assertEqual(permutations, expected_perms)

    def test_L_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move L."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('L')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 53,
            3: 50,
            6: 47,
            18: 0,
            21: 3,
            24: 6,
            27: 18,
            30: 21,
            33: 24,
            36: 42,
            37: 39,
            38: 36,
            39: 43,
            41: 37,
            42: 44,
            43: 41,
            44: 38,
            47: 33,
            50: 30,
            53: 27,
        }

        self.assertEqual(permutations, expected_perms)

    def test_L_prime_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move L'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("L'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 18,
            3: 21,
            6: 24,
            18: 27,
            21: 30,
            24: 33,
            27: 53,
            30: 50,
            33: 47,
            36: 38,
            37: 41,
            38: 44,
            39: 37,
            41: 43,
            42: 36,
            43: 39,
            44: 42,
            47: 6,
            50: 3,
            53: 0,
        }

        self.assertEqual(permutations, expected_perms)

    def test_L_2_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move L2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('L2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 27,
            3: 30,
            6: 33,
            18: 53,
            21: 50,
            24: 47,
            27: 0,
            30: 3,
            33: 6,
            36: 44,
            37: 43,
            38: 42,
            39: 41,
            41: 39,
            42: 38,
            43: 37,
            44: 36,
            47: 24,
            50: 21,
            53: 18,
        }

        self.assertEqual(permutations, expected_perms)

    def test_B_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move B."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('B')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 11,
            1: 14,
            2: 17,
            11: 35,
            14: 34,
            17: 33,
            33: 36,
            34: 39,
            35: 42,
            36: 2,
            39: 1,
            42: 0,
            45: 51,
            46: 48,
            47: 45,
            48: 52,
            50: 46,
            51: 53,
            52: 50,
            53: 47,
        }

        self.assertEqual(permutations, expected_perms)

    def test_B_prime_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move B'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("B'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 42,
            1: 39,
            2: 36,
            11: 0,
            14: 1,
            17: 2,
            33: 17,
            34: 14,
            35: 11,
            36: 33,
            39: 34,
            42: 35,
            45: 47,
            46: 50,
            47: 53,
            48: 46,
            50: 52,
            51: 45,
            52: 48,
            53: 51,
        }

        self.assertEqual(permutations, expected_perms)

    def test_B_2_move_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move B2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('B2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 35,
            1: 34,
            2: 33,
            11: 42,
            14: 39,
            17: 36,
            33: 2,
            34: 1,
            35: 0,
            36: 17,
            39: 14,
            42: 11,
            45: 53,
            46: 52,
            47: 51,
            48: 50,
            50: 48,
            51: 47,
            52: 46,
            53: 45,
        }

        self.assertEqual(permutations, expected_perms)

    def test_x_rotation_explicit_permutations(self) -> None:
        """Test explicit permutations for move x."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('x')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 18,
            1: 19,
            2: 20,
            3: 21,
            4: 22,
            5: 23,
            6: 24,
            7: 25,
            8: 26,
            9: 15,
            10: 12,
            11: 9,
            12: 16,
            14: 10,
            15: 17,
            16: 14,
            17: 11,
            18: 27,
            19: 28,
            20: 29,
            21: 30,
            22: 31,
            23: 32,
            24: 33,
            25: 34,
            26: 35,
            27: 53,
            28: 52,
            29: 51,
            30: 50,
            31: 49,
            32: 48,
            33: 47,
            34: 46,
            35: 45,
            36: 38,
            37: 41,
            38: 44,
            39: 37,
            41: 43,
            42: 36,
            43: 39,
            44: 42,
            45: 8,
            46: 7,
            47: 6,
            48: 5,
            49: 4,
            50: 3,
            51: 2,
            52: 1,
            53: 0,
        }

        self.assertEqual(permutations, expected_perms)

    def test_x_prime_rotation_explicit_permutations(self) -> None:
        """Test explicit permutations for move x'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("x'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 53,
            1: 52,
            2: 51,
            3: 50,
            4: 49,
            5: 48,
            6: 47,
            7: 46,
            8: 45,
            9: 11,
            10: 14,
            11: 17,
            12: 10,
            14: 16,
            15: 9,
            16: 12,
            17: 15,
            18: 0,
            19: 1,
            20: 2,
            21: 3,
            22: 4,
            23: 5,
            24: 6,
            25: 7,
            26: 8,
            27: 18,
            28: 19,
            29: 20,
            30: 21,
            31: 22,
            32: 23,
            33: 24,
            34: 25,
            35: 26,
            36: 42,
            37: 39,
            38: 36,
            39: 43,
            41: 37,
            42: 44,
            43: 41,
            44: 38,
            45: 35,
            46: 34,
            47: 33,
            48: 32,
            49: 31,
            50: 30,
            51: 29,
            52: 28,
            53: 27,
        }

        self.assertEqual(permutations, expected_perms)

    def test_x_2_rotation_explicit_permutations(self) -> None:
        """Test explicit permutations for move x2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('x2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 27,
            1: 28,
            2: 29,
            3: 30,
            4: 31,
            5: 32,
            6: 33,
            7: 34,
            8: 35,
            9: 17,
            10: 16,
            11: 15,
            12: 14,
            14: 12,
            15: 11,
            16: 10,
            17: 9,
            18: 53,
            19: 52,
            20: 51,
            21: 50,
            22: 49,
            23: 48,
            24: 47,
            25: 46,
            26: 45,
            27: 0,
            28: 1,
            29: 2,
            30: 3,
            31: 4,
            32: 5,
            33: 6,
            34: 7,
            35: 8,
            36: 44,
            37: 43,
            38: 42,
            39: 41,
            41: 39,
            42: 38,
            43: 37,
            44: 36,
            45: 26,
            46: 25,
            47: 24,
            48: 23,
            49: 22,
            50: 21,
            51: 20,
            52: 19,
            53: 18,
        }

        self.assertEqual(permutations, expected_perms)

    def test_y_rotation_explicit_permutations(self) -> None:
        """Test explicit permutations for move y."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('y')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 6,
            1: 3,
            2: 0,
            3: 7,
            5: 1,
            6: 8,
            7: 5,
            8: 2,
            9: 45,
            10: 46,
            11: 47,
            12: 48,
            13: 49,
            14: 50,
            15: 51,
            16: 52,
            17: 53,
            18: 9,
            19: 10,
            20: 11,
            21: 12,
            22: 13,
            23: 14,
            24: 15,
            25: 16,
            26: 17,
            27: 29,
            28: 32,
            29: 35,
            30: 28,
            32: 34,
            33: 27,
            34: 30,
            35: 33,
            36: 18,
            37: 19,
            38: 20,
            39: 21,
            40: 22,
            41: 23,
            42: 24,
            43: 25,
            44: 26,
            45: 36,
            46: 37,
            47: 38,
            48: 39,
            49: 40,
            50: 41,
            51: 42,
            52: 43,
            53: 44,
        }

        self.assertEqual(permutations, expected_perms)

    def test_y_prime_rotation_explicit_permutations(self) -> None:
        """Test explicit permutations for move y'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("y'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 2,
            1: 5,
            2: 8,
            3: 1,
            5: 7,
            6: 0,
            7: 3,
            8: 6,
            9: 18,
            10: 19,
            11: 20,
            12: 21,
            13: 22,
            14: 23,
            15: 24,
            16: 25,
            17: 26,
            18: 36,
            19: 37,
            20: 38,
            21: 39,
            22: 40,
            23: 41,
            24: 42,
            25: 43,
            26: 44,
            27: 33,
            28: 30,
            29: 27,
            30: 34,
            32: 28,
            33: 35,
            34: 32,
            35: 29,
            36: 45,
            37: 46,
            38: 47,
            39: 48,
            40: 49,
            41: 50,
            42: 51,
            43: 52,
            44: 53,
            45: 9,
            46: 10,
            47: 11,
            48: 12,
            49: 13,
            50: 14,
            51: 15,
            52: 16,
            53: 17,
        }

        self.assertEqual(permutations, expected_perms)

    def test_y_2_rotation_explicit_permutations(self) -> None:
        """Test explicit permutations for move y2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('y2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 8,
            1: 7,
            2: 6,
            3: 5,
            5: 3,
            6: 2,
            7: 1,
            8: 0,
            9: 36,
            10: 37,
            11: 38,
            12: 39,
            13: 40,
            14: 41,
            15: 42,
            16: 43,
            17: 44,
            18: 45,
            19: 46,
            20: 47,
            21: 48,
            22: 49,
            23: 50,
            24: 51,
            25: 52,
            26: 53,
            27: 35,
            28: 34,
            29: 33,
            30: 32,
            32: 30,
            33: 29,
            34: 28,
            35: 27,
            36: 9,
            37: 10,
            38: 11,
            39: 12,
            40: 13,
            41: 14,
            42: 15,
            43: 16,
            44: 17,
            45: 18,
            46: 19,
            47: 20,
            48: 21,
            49: 22,
            50: 23,
            51: 24,
            52: 25,
            53: 26,
        }

        self.assertEqual(permutations, expected_perms)

    def test_z_rotation_explicit_permutations(self) -> None:
        """Test explicit permutations for move z."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('z')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 42,
            1: 39,
            2: 36,
            3: 43,
            4: 40,
            5: 37,
            6: 44,
            7: 41,
            8: 38,
            9: 6,
            10: 3,
            11: 0,
            12: 7,
            13: 4,
            14: 1,
            15: 8,
            16: 5,
            17: 2,
            18: 24,
            19: 21,
            20: 18,
            21: 25,
            23: 19,
            24: 26,
            25: 23,
            26: 20,
            27: 15,
            28: 12,
            29: 9,
            30: 16,
            31: 13,
            32: 10,
            33: 17,
            34: 14,
            35: 11,
            36: 33,
            37: 30,
            38: 27,
            39: 34,
            40: 31,
            41: 28,
            42: 35,
            43: 32,
            44: 29,
            45: 47,
            46: 50,
            47: 53,
            48: 46,
            50: 52,
            51: 45,
            52: 48,
            53: 51,
        }

        self.assertEqual(permutations, expected_perms)

    def test_z_prime_rotation_explicit_permutations(self) -> None:
        """Test explicit permutations for move z'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("z'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 11,
            1: 14,
            2: 17,
            3: 10,
            4: 13,
            5: 16,
            6: 9,
            7: 12,
            8: 15,
            9: 29,
            10: 32,
            11: 35,
            12: 28,
            13: 31,
            14: 34,
            15: 27,
            16: 30,
            17: 33,
            18: 20,
            19: 23,
            20: 26,
            21: 19,
            23: 25,
            24: 18,
            25: 21,
            26: 24,
            27: 38,
            28: 41,
            29: 44,
            30: 37,
            31: 40,
            32: 43,
            33: 36,
            34: 39,
            35: 42,
            36: 2,
            37: 5,
            38: 8,
            39: 1,
            40: 4,
            41: 7,
            42: 0,
            43: 3,
            44: 6,
            45: 51,
            46: 48,
            47: 45,
            48: 52,
            50: 46,
            51: 53,
            52: 50,
            53: 47,
        }

        self.assertEqual(permutations, expected_perms)

    def test_z_2_rotation_explicit_permutations(self) -> None:
        """Test explicit permutations for move z2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('z2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 35,
            1: 34,
            2: 33,
            3: 32,
            4: 31,
            5: 30,
            6: 29,
            7: 28,
            8: 27,
            9: 44,
            10: 43,
            11: 42,
            12: 41,
            13: 40,
            14: 39,
            15: 38,
            16: 37,
            17: 36,
            18: 26,
            19: 25,
            20: 24,
            21: 23,
            23: 21,
            24: 20,
            25: 19,
            26: 18,
            27: 8,
            28: 7,
            29: 6,
            30: 5,
            31: 4,
            32: 3,
            33: 2,
            34: 1,
            35: 0,
            36: 17,
            37: 16,
            38: 15,
            39: 14,
            40: 13,
            41: 12,
            42: 11,
            43: 10,
            44: 9,
            45: 53,
            46: 52,
            47: 51,
            48: 50,
            50: 48,
            51: 47,
            52: 46,
            53: 45,
        }

        self.assertEqual(permutations, expected_perms)

    def test_M_slice_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move M."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('M')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            1: 52,
            4: 49,
            7: 46,
            19: 1,
            22: 4,
            25: 7,
            28: 19,
            31: 22,
            34: 25,
            46: 34,
            49: 31,
            52: 28,
        }

        self.assertEqual(permutations, expected_perms)

    def test_M_prime_slice_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move M'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("M'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            1: 19,
            4: 22,
            7: 25,
            19: 28,
            22: 31,
            25: 34,
            28: 52,
            31: 49,
            34: 46,
            46: 7,
            49: 4,
            52: 1,
        }

        self.assertEqual(permutations, expected_perms)

    def test_M_2_slice_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move M2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('M2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            1: 28,
            4: 31,
            7: 34,
            19: 52,
            22: 49,
            25: 46,
            28: 1,
            31: 4,
            34: 7,
            46: 25,
            49: 22,
            52: 19,
        }

        self.assertEqual(permutations, expected_perms)

    def test_E_slice_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move E."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('E')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            12: 21,
            13: 22,
            14: 23,
            21: 39,
            22: 40,
            23: 41,
            39: 48,
            40: 49,
            41: 50,
            48: 12,
            49: 13,
            50: 14,
        }

        self.assertEqual(permutations, expected_perms)

    def test_E_prime_slice_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move E'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("E'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            12: 48,
            13: 49,
            14: 50,
            21: 12,
            22: 13,
            23: 14,
            39: 21,
            40: 22,
            41: 23,
            48: 39,
            49: 40,
            50: 41,
        }

        self.assertEqual(permutations, expected_perms)

    def test_E_2_slice_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move E2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('E2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            12: 39,
            13: 40,
            14: 41,
            21: 48,
            22: 49,
            23: 50,
            39: 12,
            40: 13,
            41: 14,
            48: 21,
            49: 22,
            50: 23,
        }

        self.assertEqual(permutations, expected_perms)

    def test_S_slice_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move S."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('S')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            3: 43,
            4: 40,
            5: 37,
            10: 3,
            13: 4,
            16: 5,
            30: 16,
            31: 13,
            32: 10,
            37: 30,
            40: 31,
            43: 32,
        }

        self.assertEqual(permutations, expected_perms)

    def test_S_prime_slice_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move S'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("S'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            3: 10,
            4: 13,
            5: 16,
            10: 32,
            13: 31,
            16: 30,
            30: 37,
            31: 40,
            32: 43,
            37: 5,
            40: 4,
            43: 3,
        }

        self.assertEqual(permutations, expected_perms)

    def test_S_2_slice_explicit_permutations(self) -> None:  # noqa: N802
        """Test explicit permutations for move S2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('S2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            3: 32,
            4: 31,
            5: 30,
            10: 43,
            13: 40,
            16: 37,
            30: 5,
            31: 4,
            32: 3,
            37: 16,
            40: 13,
            43: 10,
        }

        self.assertEqual(permutations, expected_perms)

    def test_u_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move u."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('u')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 6,
            1: 3,
            2: 0,
            3: 7,
            5: 1,
            6: 8,
            7: 5,
            8: 2,
            9: 45,
            10: 46,
            11: 47,
            12: 48,
            13: 49,
            14: 50,
            18: 9,
            19: 10,
            20: 11,
            21: 12,
            22: 13,
            23: 14,
            36: 18,
            37: 19,
            38: 20,
            39: 21,
            40: 22,
            41: 23,
            45: 36,
            46: 37,
            47: 38,
            48: 39,
            49: 40,
            50: 41,
        }

        self.assertEqual(permutations, expected_perms)

    def test_u_prime_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move u'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("u'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 2,
            1: 5,
            2: 8,
            3: 1,
            5: 7,
            6: 0,
            7: 3,
            8: 6,
            9: 18,
            10: 19,
            11: 20,
            12: 21,
            13: 22,
            14: 23,
            18: 36,
            19: 37,
            20: 38,
            21: 39,
            22: 40,
            23: 41,
            36: 45,
            37: 46,
            38: 47,
            39: 48,
            40: 49,
            41: 50,
            45: 9,
            46: 10,
            47: 11,
            48: 12,
            49: 13,
            50: 14,
        }

        self.assertEqual(permutations, expected_perms)

    def test_u_2_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move u2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('u2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 8,
            1: 7,
            2: 6,
            3: 5,
            5: 3,
            6: 2,
            7: 1,
            8: 0,
            9: 36,
            10: 37,
            11: 38,
            12: 39,
            13: 40,
            14: 41,
            18: 45,
            19: 46,
            20: 47,
            21: 48,
            22: 49,
            23: 50,
            36: 9,
            37: 10,
            38: 11,
            39: 12,
            40: 13,
            41: 14,
            45: 18,
            46: 19,
            47: 20,
            48: 21,
            49: 22,
            50: 23,
        }

        self.assertEqual(permutations, expected_perms)

    def test_r_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move r."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('r')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            1: 19,
            2: 20,
            4: 22,
            5: 23,
            7: 25,
            8: 26,
            9: 15,
            10: 12,
            11: 9,
            12: 16,
            14: 10,
            15: 17,
            16: 14,
            17: 11,
            19: 28,
            20: 29,
            22: 31,
            23: 32,
            25: 34,
            26: 35,
            28: 52,
            29: 51,
            31: 49,
            32: 48,
            34: 46,
            35: 45,
            45: 8,
            46: 7,
            48: 5,
            49: 4,
            51: 2,
            52: 1,
        }

        self.assertEqual(permutations, expected_perms)

    def test_r_prime_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move r'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("r'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            1: 52,
            2: 51,
            4: 49,
            5: 48,
            7: 46,
            8: 45,
            9: 11,
            10: 14,
            11: 17,
            12: 10,
            14: 16,
            15: 9,
            16: 12,
            17: 15,
            19: 1,
            20: 2,
            22: 4,
            23: 5,
            25: 7,
            26: 8,
            28: 19,
            29: 20,
            31: 22,
            32: 23,
            34: 25,
            35: 26,
            45: 35,
            46: 34,
            48: 32,
            49: 31,
            51: 29,
            52: 28,
        }

        self.assertEqual(permutations, expected_perms)

    def test_r_2_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move r2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('r2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            1: 28,
            2: 29,
            4: 31,
            5: 32,
            7: 34,
            8: 35,
            9: 17,
            10: 16,
            11: 15,
            12: 14,
            14: 12,
            15: 11,
            16: 10,
            17: 9,
            19: 52,
            20: 51,
            22: 49,
            23: 48,
            25: 46,
            26: 45,
            28: 1,
            29: 2,
            31: 4,
            32: 5,
            34: 7,
            35: 8,
            45: 26,
            46: 25,
            48: 23,
            49: 22,
            51: 20,
            52: 19,
        }

        self.assertEqual(permutations, expected_perms)

    def test_f_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move f."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('f')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            3: 43,
            4: 40,
            5: 37,
            6: 44,
            7: 41,
            8: 38,
            9: 6,
            10: 3,
            12: 7,
            13: 4,
            15: 8,
            16: 5,
            18: 24,
            19: 21,
            20: 18,
            21: 25,
            23: 19,
            24: 26,
            25: 23,
            26: 20,
            27: 15,
            28: 12,
            29: 9,
            30: 16,
            31: 13,
            32: 10,
            37: 30,
            38: 27,
            40: 31,
            41: 28,
            43: 32,
            44: 29,
        }

        self.assertEqual(permutations, expected_perms)

    def test_f_prime_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move f'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("f'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            3: 10,
            4: 13,
            5: 16,
            6: 9,
            7: 12,
            8: 15,
            9: 29,
            10: 32,
            12: 28,
            13: 31,
            15: 27,
            16: 30,
            18: 20,
            19: 23,
            20: 26,
            21: 19,
            23: 25,
            24: 18,
            25: 21,
            26: 24,
            27: 38,
            28: 41,
            29: 44,
            30: 37,
            31: 40,
            32: 43,
            37: 5,
            38: 8,
            40: 4,
            41: 7,
            43: 3,
            44: 6,
        }

        self.assertEqual(permutations, expected_perms)

    def test_f_2_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move f2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('f2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            3: 32,
            4: 31,
            5: 30,
            6: 29,
            7: 28,
            8: 27,
            9: 44,
            10: 43,
            12: 41,
            13: 40,
            15: 38,
            16: 37,
            18: 26,
            19: 25,
            20: 24,
            21: 23,
            23: 21,
            24: 20,
            25: 19,
            26: 18,
            27: 8,
            28: 7,
            29: 6,
            30: 5,
            31: 4,
            32: 3,
            37: 16,
            38: 15,
            40: 13,
            41: 12,
            43: 10,
            44: 9,
        }

        self.assertEqual(permutations, expected_perms)

    def test_d_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move d."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('d')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            12: 21,
            13: 22,
            14: 23,
            15: 24,
            16: 25,
            17: 26,
            21: 39,
            22: 40,
            23: 41,
            24: 42,
            25: 43,
            26: 44,
            27: 33,
            28: 30,
            29: 27,
            30: 34,
            32: 28,
            33: 35,
            34: 32,
            35: 29,
            39: 48,
            40: 49,
            41: 50,
            42: 51,
            43: 52,
            44: 53,
            48: 12,
            49: 13,
            50: 14,
            51: 15,
            52: 16,
            53: 17,
        }

        self.assertEqual(permutations, expected_perms)

    def test_d_prime_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move d'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("d'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            12: 48,
            13: 49,
            14: 50,
            15: 51,
            16: 52,
            17: 53,
            21: 12,
            22: 13,
            23: 14,
            24: 15,
            25: 16,
            26: 17,
            27: 29,
            28: 32,
            29: 35,
            30: 28,
            32: 34,
            33: 27,
            34: 30,
            35: 33,
            39: 21,
            40: 22,
            41: 23,
            42: 24,
            43: 25,
            44: 26,
            48: 39,
            49: 40,
            50: 41,
            51: 42,
            52: 43,
            53: 44,
        }

        self.assertEqual(permutations, expected_perms)

    def test_d_2_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move d2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('d2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            12: 39,
            13: 40,
            14: 41,
            15: 42,
            16: 43,
            17: 44,
            21: 48,
            22: 49,
            23: 50,
            24: 51,
            25: 52,
            26: 53,
            27: 35,
            28: 34,
            29: 33,
            30: 32,
            32: 30,
            33: 29,
            34: 28,
            35: 27,
            39: 12,
            40: 13,
            41: 14,
            42: 15,
            43: 16,
            44: 17,
            48: 21,
            49: 22,
            50: 23,
            51: 24,
            52: 25,
            53: 26,
        }

        self.assertEqual(permutations, expected_perms)

    def test_l_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move l."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('l')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 53,
            1: 52,
            3: 50,
            4: 49,
            6: 47,
            7: 46,
            18: 0,
            19: 1,
            21: 3,
            22: 4,
            24: 6,
            25: 7,
            27: 18,
            28: 19,
            30: 21,
            31: 22,
            33: 24,
            34: 25,
            36: 42,
            37: 39,
            38: 36,
            39: 43,
            41: 37,
            42: 44,
            43: 41,
            44: 38,
            46: 34,
            47: 33,
            49: 31,
            50: 30,
            52: 28,
            53: 27,
        }

        self.assertEqual(permutations, expected_perms)

    def test_l_prime_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move l'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("l'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 18,
            1: 19,
            3: 21,
            4: 22,
            6: 24,
            7: 25,
            18: 27,
            19: 28,
            21: 30,
            22: 31,
            24: 33,
            25: 34,
            27: 53,
            28: 52,
            30: 50,
            31: 49,
            33: 47,
            34: 46,
            36: 38,
            37: 41,
            38: 44,
            39: 37,
            41: 43,
            42: 36,
            43: 39,
            44: 42,
            46: 7,
            47: 6,
            49: 4,
            50: 3,
            52: 1,
            53: 0,
        }

        self.assertEqual(permutations, expected_perms)

    def test_l_2_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move l2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('l2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 27,
            1: 28,
            3: 30,
            4: 31,
            6: 33,
            7: 34,
            18: 53,
            19: 52,
            21: 50,
            22: 49,
            24: 47,
            25: 46,
            27: 0,
            28: 1,
            30: 3,
            31: 4,
            33: 6,
            34: 7,
            36: 44,
            37: 43,
            38: 42,
            39: 41,
            41: 39,
            42: 38,
            43: 37,
            44: 36,
            46: 25,
            47: 24,
            49: 22,
            50: 21,
            52: 19,
            53: 18,
        }

        self.assertEqual(permutations, expected_perms)

    def test_b_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move b."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('b')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 11,
            1: 14,
            2: 17,
            3: 10,
            4: 13,
            5: 16,
            10: 32,
            11: 35,
            13: 31,
            14: 34,
            16: 30,
            17: 33,
            30: 37,
            31: 40,
            32: 43,
            33: 36,
            34: 39,
            35: 42,
            36: 2,
            37: 5,
            39: 1,
            40: 4,
            42: 0,
            43: 3,
            45: 51,
            46: 48,
            47: 45,
            48: 52,
            50: 46,
            51: 53,
            52: 50,
            53: 47,
        }

        self.assertEqual(permutations, expected_perms)

    def test_b_prime_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move b'."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate("b'")
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 42,
            1: 39,
            2: 36,
            3: 43,
            4: 40,
            5: 37,
            10: 3,
            11: 0,
            13: 4,
            14: 1,
            16: 5,
            17: 2,
            30: 16,
            31: 13,
            32: 10,
            33: 17,
            34: 14,
            35: 11,
            36: 33,
            37: 30,
            39: 34,
            40: 31,
            42: 35,
            43: 32,
            45: 47,
            46: 50,
            47: 53,
            48: 46,
            50: 52,
            51: 45,
            52: 48,
            53: 51,
        }

        self.assertEqual(permutations, expected_perms)

    def test_b_2_wide_explicit_permutations(self) -> None:
        """Test explicit permutations for move b2."""
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('b2')
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        expected_perms = {
            0: 35,
            1: 34,
            2: 33,
            3: 32,
            4: 31,
            5: 30,
            10: 43,
            11: 42,
            13: 40,
            14: 39,
            16: 37,
            17: 36,
            30: 5,
            31: 4,
            32: 3,
            33: 2,
            34: 1,
            35: 0,
            36: 17,
            37: 16,
            39: 14,
            40: 13,
            42: 11,
            43: 10,
            45: 53,
            46: 52,
            47: 51,
            48: 50,
            50: 48,
            51: 47,
            52: 46,
            53: 45,
        }

        self.assertEqual(permutations, expected_perms)



import time
import unittest
from collections import Counter
from statistics import mean
from statistics import stdev

from cubing_algs.algorithm import Algorithm
from cubing_algs.constants import FACE_ORDER
from cubing_algs.constants import OPPOSITE_FACES
from cubing_algs.scrambler import build_cube_move_set
from cubing_algs.scrambler import is_valid_next_move
from cubing_algs.scrambler import random_moves
from cubing_algs.scrambler import scramble
from cubing_algs.scrambler import scramble_easy_cross
from cubing_algs.vcube import VCube


class TestValidNextMove(unittest.TestCase):

    def test_is_valid_next_move_valid(self) -> None:
        """Test that valid next moves are recognized."""
        self.assertTrue(is_valid_next_move('F', 'R'))
        self.assertTrue(is_valid_next_move("F'", 'R'))
        self.assertTrue(is_valid_next_move('F2', "R'"))

    def test_is_valid_next_move_invalid_same_face(self) -> None:
        """Test that moves on the same face are invalid."""
        self.assertFalse(is_valid_next_move('F', 'F'))
        self.assertFalse(is_valid_next_move('F', "F'"))
        self.assertFalse(is_valid_next_move('F2', 'F'))

    def test_is_valid_next_move_invalid_none(self) -> None:
        """Test that moves does not matche."""
        self.assertFalse(is_valid_next_move('', 'F'))
        self.assertFalse(is_valid_next_move('F', ''))
        self.assertFalse(is_valid_next_move('Z', 'F'))

    def test_is_valid_next_move_invalid_opposite_faces(self) -> None:
        """Test that moves on opposite faces are invalid."""
        self.assertFalse(is_valid_next_move('F', 'B'))
        self.assertFalse(is_valid_next_move('R', 'L'))
        self.assertFalse(is_valid_next_move('U', 'D'))

    def test_is_valid_next_move_with_modifiers(self) -> None:
        """Test that modifiers don't affect face validation."""
        self.assertFalse(is_valid_next_move('Fw', 'F'))
        self.assertFalse(is_valid_next_move('Fw', 'B'))


class TestCubeMoveSet(unittest.TestCase):
    maxDiff = None

    def test_build_cube_move_set_2x2x2(self) -> None:
        self.assertEqual(
            build_cube_move_set(2),
            [
                'R', "R'", 'R2',
                'F', "F'", 'F2',
                'U', "U'", 'U2',
                'L', "L'", 'L2',
                'B', "B'", 'B2',
                'D', "D'", 'D2',
            ],
        )

    def test_build_cube_move_set_2x2x2_inner_layers(self) -> None:
        self.assertEqual(
            build_cube_move_set(2, inner_layers=True),
            [
                'R', "R'", 'R2',
                'F', "F'", 'F2',
                'U', "U'", 'U2',
                'L', "L'", 'L2',
                'B', "B'", 'B2',
                'D', "D'", 'D2',
            ],
        )

    def test_build_cube_move_set_3x3x3(self) -> None:
        self.assertEqual(
            build_cube_move_set(3),
            [
                'R', "R'", 'R2',
                'F', "F'", 'F2',
                'U', "U'", 'U2',
                'L', "L'", 'L2',
                'B', "B'", 'B2',
                'D', "D'", 'D2',
            ],
        )

    def test_build_cube_move_set_3x3x3_inner_layers(self) -> None:
        self.assertEqual(
            build_cube_move_set(3, inner_layers=True),
            [
                'R', "R'", 'R2',
                'F', "F'", 'F2',
                'U', "U'", 'U2',
                'L', "L'", 'L2',
                'B', "B'", 'B2',
                'D', "D'", 'D2',
            ],
        )

    def test_build_cube_move_set_4x4x4(self) -> None:
        self.assertEqual(
            build_cube_move_set(4),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                'L', "L'", 'L2',
                'B', "B'", 'B2',
                'D', "D'", 'D2',
            ],
        )

    def test_build_cube_move_set_4x4x4_left_handed(self) -> None:
        self.assertEqual(
            build_cube_move_set(4, right_handed=False),
            [
                'R', "R'", 'R2',
                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                'B', "B'", 'B2',
                'D', "D'", 'D2',
            ],
        )

    def test_build_cube_move_set_4x4x4_inner_layers(self) -> None:
        self.assertEqual(
            build_cube_move_set(4, inner_layers=True),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                '2R', "2R'", '2R2',
                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '2F', "2F'", '2F2',
                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '2U', "2U'", '2U2',
                'L', "L'", 'L2',
                '2L', "2L'", '2L2',
                'B', "B'", 'B2',
                '2B', "2B'", '2B2',
                'D', "D'", 'D2',
                '2D', "2D'", '2D2',
            ],
        )

    def test_build_cube_move_set_4x4x4_inner_layers_left_handed(self) -> None:
        self.assertEqual(
            build_cube_move_set(4, inner_layers=True, right_handed=False),
            [
                'R', "R'", 'R2',
                '2R', "2R'", '2R2',
                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '2F', "2F'", '2F2',
                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '2U', "2U'", '2U2',
                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                '2L', "2L'", '2L2',
                'B', "B'", 'B2',
                '2B', "2B'", '2B2',
                'D', "D'", 'D2',
                '2D', "2D'", '2D2',
            ],
        )

    def test_build_cube_move_set_5x5x5(self) -> None:
        self.assertEqual(
            build_cube_move_set(5),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',
                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
            ],
        )

    def test_build_cube_move_set_5x5x5_left_handed(self) -> None:
        self.assertEqual(
            build_cube_move_set(5, right_handed=False),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',
                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
            ],
        )

    def test_build_cube_move_set_5x5x5_inner_layers(self) -> None:
        self.assertEqual(
            build_cube_move_set(5, inner_layers=True),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                '2R', "2R'", '2R2', '3R', "3R'", '3R2',

                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '2F', "2F'", '2F2', '3F', "3F'", '3F2',

                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '2U', "2U'", '2U2', '3U', "3U'", '3U2',

                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                '2L', "2L'", '2L2',

                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',
                '2B', "2B'", '2B2',

                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
                '2D', "2D'", '2D2',
            ],
        )

    def test_build_cube_move_set_5x5x5_inner_layers_left_handed(self) -> None:
        self.assertEqual(
            build_cube_move_set(5, inner_layers=True, right_handed=False),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                '2R', "2R'", '2R2',

                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '2F', "2F'", '2F2', '3F', "3F'", '3F2',

                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '2U', "2U'", '2U2', '3U', "3U'", '3U2',

                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                '2L', "2L'", '2L2', '3L', "3L'", '3L2',

                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',
                '2B', "2B'", '2B2',

                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
                '2D', "2D'", '2D2',
            ],
        )

    def test_build_cube_move_set_6x6x6(self) -> None:
        self.assertEqual(
            build_cube_move_set(6),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                '3Rw', "3Rw'", '3Rw2',

                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '3Fw', "3Fw'", '3Fw2',

                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '3Uw', "3Uw'", '3Uw2',

                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',

                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',

                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
            ],
        )

    def test_build_cube_move_set_6x6x6_inner_layers(self) -> None:
        self.assertEqual(
            build_cube_move_set(6, inner_layers=True),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                '3Rw', "3Rw'", '3Rw2',
                '2R', "2R'", '2R2',
                '3R', "3R'", '3R2',

                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '3Fw', "3Fw'", '3Fw2',
                '2F', "2F'", '2F2',
                '3F', "3F'", '3F2',

                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '3Uw', "3Uw'", '3Uw2',
                '2U', "2U'", '2U2',
                '3U', "3U'", '3U2',

                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                '2L', "2L'", '2L2',
                '3L', "3L'", '3L2',

                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',
                '2B', "2B'", '2B2',
                '3B', "3B'", '3B2',

                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
                '2D', "2D'", '2D2',
                '3D', "3D'", '3D2',
            ],
        )

    def test_build_cube_move_set_6x6x6_left_handed(self) -> None:
        self.assertEqual(
            build_cube_move_set(6, right_handed=False),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',

                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '3Fw', "3Fw'", '3Fw2',

                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '3Uw', "3Uw'", '3Uw2',

                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                '3Lw', "3Lw'", '3Lw2',

                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',

                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
            ],
        )

    def test_build_cube_move_set_6x6x6_inner_layers_left_handed(self) -> None:
        self.assertEqual(
            build_cube_move_set(6, inner_layers=True, right_handed=False),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                '2R', "2R'", '2R2',
                '3R', "3R'", '3R2',

                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '3Fw', "3Fw'", '3Fw2',
                '2F', "2F'", '2F2',
                '3F', "3F'", '3F2',

                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '3Uw', "3Uw'", '3Uw2',
                '2U', "2U'", '2U2',
                '3U', "3U'", '3U2',

                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                '3Lw', "3Lw'", '3Lw2',
                '2L', "2L'", '2L2',
                '3L', "3L'", '3L2',

                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',
                '2B', "2B'", '2B2',
                '3B', "3B'", '3B2',

                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
                '2D', "2D'", '2D2',
                '3D', "3D'", '3D2',
            ],
        )

    def test_build_cube_move_set_7x7x7(self) -> None:
        self.assertEqual(
            build_cube_move_set(7),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                '3Rw', "3Rw'", '3Rw2',

                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '3Fw', "3Fw'", '3Fw2',

                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '3Uw', "3Uw'", '3Uw2',

                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                '3Lw', "3Lw'", '3Lw2',

                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',
                '3Bw', "3Bw'", '3Bw2',

                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
                '3Dw', "3Dw'", '3Dw2',
            ],
        )

    def test_build_cube_move_set_7x7x7_inner_layers(self) -> None:
        self.assertEqual(
            build_cube_move_set(7, inner_layers=True),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                '3Rw', "3Rw'", '3Rw2',
                '2R', "2R'", '2R2',
                '3R', "3R'", '3R2',
                '4R', "4R'", '4R2',

                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '3Fw', "3Fw'", '3Fw2',
                '2F', "2F'", '2F2',
                '3F', "3F'", '3F2',
                '4F', "4F'", '4F2',

                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '3Uw', "3Uw'", '3Uw2',
                '2U', "2U'", '2U2',
                '3U', "3U'", '3U2',
                '4U', "4U'", '4U2',

                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                '3Lw', "3Lw'", '3Lw2',
                '2L', "2L'", '2L2',
                '3L', "3L'", '3L2',

                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',
                '3Bw', "3Bw'", '3Bw2',
                '2B', "2B'", '2B2',
                '3B', "3B'", '3B2',

                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
                '3Dw', "3Dw'", '3Dw2',
                '2D', "2D'", '2D2',
                '3D', "3D'", '3D2',
            ],
        )

    def test_build_cube_move_set_8x8x8(self) -> None:
        self.assertEqual(
            build_cube_move_set(8),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                '3Rw', "3Rw'", '3Rw2',
                '4Rw', "4Rw'", '4Rw2',

                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '3Fw', "3Fw'", '3Fw2',
                '4Fw', "4Fw'", '4Fw2',

                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '3Uw', "3Uw'", '3Uw2',
                '4Uw', "4Uw'", '4Uw2',

                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                '3Lw', "3Lw'", '3Lw2',

                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',
                '3Bw', "3Bw'", '3Bw2',

                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
                '3Dw', "3Dw'", '3Dw2',
            ],
        )

    def test_build_cube_move_set_8x8x8_inner_layers(self) -> None:
        self.assertEqual(
            build_cube_move_set(8, inner_layers=True),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                '3Rw', "3Rw'", '3Rw2',
                '4Rw', "4Rw'", '4Rw2',
                '2R', "2R'", '2R2',
                '3R', "3R'", '3R2',
                '4R', "4R'", '4R2',

                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '3Fw', "3Fw'", '3Fw2',
                '4Fw', "4Fw'", '4Fw2',
                '2F', "2F'", '2F2',
                '3F', "3F'", '3F2',
                '4F', "4F'", '4F2',

                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '3Uw', "3Uw'", '3Uw2',
                '4Uw', "4Uw'", '4Uw2',
                '2U', "2U'", '2U2',
                '3U', "3U'", '3U2',
                '4U', "4U'", '4U2',

                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                '3Lw', "3Lw'", '3Lw2',
                '2L', "2L'", '2L2',
                '3L', "3L'", '3L2',
                '4L', "4L'", '4L2',

                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',
                '3Bw', "3Bw'", '3Bw2',
                '2B', "2B'", '2B2',
                '3B', "3B'", '3B2',
                '4B', "4B'", '4B2',

                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
                '3Dw', "3Dw'", '3Dw2',
                '2D', "2D'", '2D2',
                '3D', "3D'", '3D2',
                '4D', "4D'", '4D2',
            ],
        )

    def test_build_cube_move_set_9x9x9(self) -> None:
        self.assertEqual(
            build_cube_move_set(9),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                '3Rw', "3Rw'", '3Rw2',
                '4Rw', "4Rw'", '4Rw2',

                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '3Fw', "3Fw'", '3Fw2',
                '4Fw', "4Fw'", '4Fw2',

                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '3Uw', "3Uw'", '3Uw2',
                '4Uw', "4Uw'", '4Uw2',

                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                '3Lw', "3Lw'", '3Lw2',
                '4Lw', "4Lw'", '4Lw2',

                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',
                '3Bw', "3Bw'", '3Bw2',
                '4Bw', "4Bw'", '4Bw2',

                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
                '3Dw', "3Dw'", '3Dw2',
                '4Dw', "4Dw'", '4Dw2',
            ],
        )

    def test_build_cube_move_set_9x9x9_inner_layers(self) -> None:
        self.assertEqual(
            build_cube_move_set(9, inner_layers=True),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                '3Rw', "3Rw'", '3Rw2',
                '4Rw', "4Rw'", '4Rw2',
                '2R', "2R'", '2R2',
                '3R', "3R'", '3R2',
                '4R', "4R'", '4R2',
                '5R', "5R'", '5R2',

                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '3Fw', "3Fw'", '3Fw2',
                '4Fw', "4Fw'", '4Fw2',
                '2F', "2F'", '2F2',
                '3F', "3F'", '3F2',
                '4F', "4F'", '4F2',
                '5F', "5F'", '5F2',

                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '3Uw', "3Uw'", '3Uw2',
                '4Uw', "4Uw'", '4Uw2',
                '2U', "2U'", '2U2',
                '3U', "3U'", '3U2',
                '4U', "4U'", '4U2',
                '5U', "5U'", '5U2',

                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                '3Lw', "3Lw'", '3Lw2',
                '4Lw', "4Lw'", '4Lw2',
                '2L', "2L'", '2L2',
                '3L', "3L'", '3L2',
                '4L', "4L'", '4L2',

                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',
                '3Bw', "3Bw'", '3Bw2',
                '4Bw', "4Bw'", '4Bw2',
                '2B', "2B'", '2B2',
                '3B', "3B'", '3B2',
                '4B', "4B'", '4B2',

                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
                '3Dw', "3Dw'", '3Dw2',
                '4Dw', "4Dw'", '4Dw2',
                '2D', "2D'", '2D2',
                '3D', "3D'", '3D2',
                '4D', "4D'", '4D2',
            ],
        )

    def test_build_big_cube_move_set_no_options(self) -> None:
        self.assertEqual(
            build_cube_move_set(6),
            [
                'R', "R'", 'R2', 'Rw', "Rw'", 'Rw2',
                '3Rw', "3Rw'", '3Rw2',

                'F', "F'", 'F2', 'Fw', "Fw'", 'Fw2',
                '3Fw', "3Fw'", '3Fw2',

                'U', "U'", 'U2', 'Uw', "Uw'", 'Uw2',
                '3Uw', "3Uw'", '3Uw2',

                'L', "L'", 'L2', 'Lw', "Lw'", 'Lw2',
                'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2',
                'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2',
            ],
        )


class TestRandomMoves(unittest.TestCase):

    def test_random_moves_2x2x2(self) -> None:
        moves = random_moves(2, ['F', 'R', 'U'], 0)

        self.assertGreaterEqual(
            len(moves), 9,
        )

        self.assertLessEqual(
            len(moves), 11,
        )

    def test_random_moves_2x2x2_iterations(self) -> None:
        moves = random_moves(2, ['F', 'R', 'U'], 5)

        self.assertEqual(
            len(moves), 5,
        )

    def test_random_moves_50x50x50(self) -> None:
        moves = random_moves(50, ['F', 'R', 'U'])

        self.assertEqual(
            len(moves), 100,
        )


class TestScramble(unittest.TestCase):

    def test_scramble_3x3x3(self) -> None:
        moves = scramble(3)

        self.assertGreaterEqual(
            len(moves), 25,
        )

        self.assertLessEqual(
            len(moves), 30,
        )

    def test_scramble_3x3x3_iterations(self) -> None:
        moves = scramble(3, 5)

        self.assertEqual(
            len(moves), 5,
        )


class TestScrambleEasyCross(unittest.TestCase):

    def test_scramble_easy_cross(self) -> None:
        moves = scramble_easy_cross()

        self.assertEqual(
            len(moves), 10,
        )
        self.assertTrue(
            'U' not in moves,
        )
        self.assertTrue(
            'D' not in moves,
        )


class TestScrambleEffectivenessByLength(unittest.TestCase):
    """
    Test suite measuring scramble effectiveness across different lengths.

    Tests randomness, state space coverage, consistency, and performance
    characteristics of scrambles at various lengths.
    """

    def setUp(self) -> None:
        """
        Set up test fixtures.
        """
        self.test_lengths = [5, 10, 15, 20, 25, 30, 50]
        self.sample_size = 100  # Number of scrambles to test per length
        self.small_sample_size = 20  # For performance tests

    @staticmethod
    def get_face_from_move(move: str) -> str:
        """
        Extract the face character from a move string.
        """
        for face in FACE_ORDER:
            if move.startswith((face, face.lower())):
                return face
        return ''

    @staticmethod
    def calculate_move_distribution(algorithm: Algorithm) -> dict[
            str, float]:
        """
        Calculate the distribution of face moves in an algorithm.
        """
        face_counts: Counter[str] = Counter()
        total_moves = len(algorithm)

        for move in algorithm:
            face = TestScrambleEffectivenessByLength.get_face_from_move(
                str(move),
            )
            if face:
                face_counts[face] += 1

        # Convert to percentages
        return {
            face: (count / total_moves) * 100
            for face, count in face_counts.items()
        }

    @staticmethod
    def calculate_state_scrambledness(algorithm: Algorithm) -> float:
        """
        Calculate how scrambled a cube becomes after applying an algorithm.

        Returns a value between 0.0 (solved) and 1.0 (maximally scrambled).
        """
        # Count how many facelets are not in their original position
        cube = VCube()
        solved_state = cube.state
        cube.rotate(algorithm)
        scrambled_state = cube.state

        different_facelets = sum(
            1
            for i, j in zip(solved_state, scrambled_state, strict=False)
            if i != j
        )

        # Normalize by total number of facelets (54 for a 3x3x3 cube)
        return different_facelets / 54.0

    @staticmethod
    def has_consecutive_same_or_opposite_faces(algorithm: Algorithm) -> bool:
        """
        Check if algorithm has consecutive moves on same or opposite faces.
        """
        moves = [str(move) for move in algorithm]

        for i in range(len(moves) - 1):
            current_face = TestScrambleEffectivenessByLength.get_face_from_move(
                moves[i],
            )
            next_face = TestScrambleEffectivenessByLength.get_face_from_move(
                moves[i + 1],
            )

            if current_face and next_face:
                # Check same face
                if current_face == next_face:
                    return True
                # Check opposite faces
                if OPPOSITE_FACES.get(current_face) == next_face:
                    return True

        return False

    def test_scramble_length_consistency(self) -> None:
        """
        Test that scrambles generate the requested length consistently.
        """
        for length in self.test_lengths:
            with self.subTest(length=length):
                for _ in range(10):  # Test multiple samples
                    scramble_alg = scramble(3, length)
                    self.assertEqual(
                        len(scramble_alg), length,
                        f'Scramble should be exactly {length} moves',
                    )

    def test_move_distribution_randomness(self) -> None:
        """
        Test that move distributions approach uniform across face types.
        """
        for length in [15, 25, 50]:  # Test representative lengths
            with self.subTest(length=length):
                all_distributions = []

                for _ in range(self.sample_size):
                    scramble_alg = scramble(3, length)
                    distribution = self.calculate_move_distribution(
                        scramble_alg,
                    )
                    all_distributions.append(distribution)

                # Calculate average distribution across all scrambles
                face_averages = {}
                for face in FACE_ORDER:
                    face_percentages = [
                        d.get(face, 0) for d in all_distributions
                    ]
                    face_averages[face] = mean(face_percentages)

                # For longer scrambles, distribution should be more uniform
                if length >= 25:
                    # Each face should appear roughly 16.67% of the time (1/6)
                    expected_percentage = 100.0 / 6
                    tolerance = 5.0  # Allow 5% deviation

                    for face, avg_percentage in face_averages.items():
                        self.assertGreater(
                            avg_percentage,
                            expected_percentage - tolerance,
                            f'Face {face} appears too rarely: '
                            f'{avg_percentage:.1f}%',
                        )
                        self.assertLess(
                            avg_percentage, expected_percentage + tolerance,
                            f'Face {face} appears too frequently: '
                            f'{avg_percentage:.1f}%',
                        )

    def test_state_space_coverage_by_length(self) -> None:
        """
        Test that longer scrambles achieve better state space coverage.
        """
        scrambledness_by_length = {}

        for length in self.test_lengths:
            scrambledness_values = []

            for _ in range(self.sample_size):
                scramble_alg = scramble(3, length)
                scrambledness = self.calculate_state_scrambledness(scramble_alg)
                scrambledness_values.append(scrambledness)

            avg_scrambledness = mean(scrambledness_values)
            scrambledness_by_length[length] = avg_scrambledness

        # Test that scrambledness generally increases with length
        lengths = sorted(scrambledness_by_length.keys())
        for i in range(len(lengths) - 1):
            current_length = lengths[i]
            next_length = lengths[i + 1]

            current_scrambledness = scrambledness_by_length[current_length]
            next_scrambledness = scrambledness_by_length[next_length]

            # Allow some variance, but general trend should be increasing
            self.assertGreaterEqual(
                next_scrambledness,
                current_scrambledness - 0.05,
                f'Scrambledness should increase with length: '
                f'{current_length}({current_scrambledness:.3f}) vs '
                f'{next_length}({next_scrambledness:.3f})',
            )

    def test_no_consecutive_invalid_moves(self) -> None:
        """
        Test that scrambles never contain consecutive invalid moves.
        """
        for length in self.test_lengths:
            with self.subTest(length=length):
                for _ in range(self.sample_size):
                    scramble_alg = scramble(3, length)
                    self.assertFalse(
                        self.has_consecutive_same_or_opposite_faces(scramble_alg),
                        f'Scramble of length {length} contains '
                        'consecutive invalid moves',
                    )

    def test_scramble_uniqueness(self) -> None:
        """
        Test that repeated scramble calls produce different results.
        """
        for length in [15, 25]:
            with self.subTest(length=length):
                scrambles = set()

                for _ in range(50):  # Generate 50 scrambles
                    scramble_alg = scramble(3, length)
                    scrambles.add(str(scramble_alg))

                # Should have high uniqueness (at least 90% unique)
                uniqueness_ratio = len(scrambles) / 50
                self.assertGreater(
                    uniqueness_ratio, 0.9,
                    f'Scrambles of length {length} not unique enough: '
                    f'{uniqueness_ratio:.2f}',
                )

    def test_very_short_scrambles(self) -> None:
        """
        Test edge cases with very short scrambles (1-5 moves).
        """
        for length in range(1, 6):
            with self.subTest(length=length):
                scramble_alg = scramble(3, length)

                # Should generate exactly the requested length
                self.assertEqual(len(scramble_alg), length)

                # Should not have invalid consecutive moves
                self.assertFalse(
                    self.has_consecutive_same_or_opposite_faces(scramble_alg),
                )

                # Should produce some scrambling effect
                if length >= 3:
                    scrambledness = self.calculate_state_scrambledness(
                        scramble_alg,
                    )
                    self.assertGreater(
                        scrambledness, 0.0,
                        'Even short scrambles should change cube state',
                    )

    def test_very_long_scrambles(self) -> None:
        """
        Test edge cases with very long scrambles (50+ moves).
        """
        long_lengths = [50, 75, 100]

        for length in long_lengths:
            with self.subTest(length=length):
                scramble_alg = scramble(3, length)

                # Should generate exactly the requested length
                self.assertEqual(len(scramble_alg), length)

                # Should not have invalid consecutive moves
                self.assertFalse(
                    self.has_consecutive_same_or_opposite_faces(scramble_alg),
                )

                # Should achieve high scrambledness
                scrambledness = self.calculate_state_scrambledness(scramble_alg)
                self.assertGreater(
                    scrambledness, 0.6,
                    'Long scrambles should achieve '
                    f'high scrambledness: {scrambledness:.3f}',
                )

    def test_scramble_performance_by_length(self) -> None:
        """
        Test performance characteristics of scramble generation.
        """
        performance_results = {}

        for length in [10, 25, 50, 100]:
            start_time = time.time()

            for _ in range(self.small_sample_size):
                scramble(3, length)

            end_time = time.time()
            avg_time_per_scramble = (
                (end_time - start_time) / self.small_sample_size
            )
            performance_results[length] = avg_time_per_scramble

        # Performance should scale reasonably with length
        for length, avg_time in performance_results.items():
            # Should generate scrambles quickly (under 10ms each)
            self.assertLess(
                avg_time, 0.01,
                'Scramble generation too slow '
                f'for length {length}: {avg_time:.4f}s',
            )

    def test_scramble_consistency_across_runs(self) -> None:
        """
        Test that scramble quality metrics are consistent across multiple runs.
        """
        length = 25  # Test with a representative length

        run_scrambledness = []
        run_uniqueness = []

        # Run multiple test batches
        for _ in range(5):
            batch_scrambledness = []
            batch_scrambles = set()

            for _ in range(20):
                scramble_alg = scramble(3, length)
                batch_scrambledness.append(
                    self.calculate_state_scrambledness(scramble_alg),
                )
                batch_scrambles.add(str(scramble_alg))

            run_scrambledness.append(mean(batch_scrambledness))
            run_uniqueness.append(len(batch_scrambles) / 20)

        # Consistency check: standard deviation should be reasonably low
        scrambledness_stdev = stdev(run_scrambledness)
        uniqueness_stdev = stdev(run_uniqueness)

        self.assertLess(
            scrambledness_stdev, 0.05,
            'Scrambledness too inconsistent across runs: '
            f'{scrambledness_stdev:.4f}')
        self.assertLess(
            uniqueness_stdev, 0.1,
            'Uniqueness too inconsistent across runs: '
            f'{uniqueness_stdev:.4f}')

    def test_move_type_distribution(self) -> None:
        """
        Test that different move types (normal, prime, double)
        are well distributed.
        """
        length = 30

        move_type_counts = {'normal': 0, 'prime': 0, 'double': 0}
        total_moves = 0

        for _ in range(self.sample_size):
            scramble_alg = scramble(3, length)

            for move in scramble_alg:
                total_moves += 1

                if move.is_counter_clockwise:
                    move_type_counts['prime'] += 1
                elif move.is_double:
                    move_type_counts['double'] += 1
                else:
                    move_type_counts['normal'] += 1

        # Convert to percentages
        move_type_percentages = {}
        for move_type, count in move_type_counts.items():
            move_type_percentages[move_type] = (count / total_moves) * 100

        # Each move type should appear roughly equally (around 33% each)
        expected_percentage = 100.0 / 3
        tolerance = 8.0  # Allow 8% deviation

        for move_type, percentage in move_type_percentages.items():
            self.assertGreater(
                percentage, expected_percentage - tolerance,
                f'Move type {move_type} appears too rarely: {percentage:.1f}%',
            )
            self.assertLess(
                percentage, expected_percentage + tolerance,
                f'Move type {move_type} appears '
                f'too frequently: {percentage:.1f}%',
            )



import unittest
from typing import Any

from cubing_algs.vcube import VCube


class TestMovePermutations(unittest.TestCase):

    @staticmethod
    def create_numbered_cube() -> VCube:
        state = ''.join([chr(ord('A') + i) for i in range(54)])
        return VCube(initial=state, check=False)

    @staticmethod
    def get_permutations(initial_state: str,
                         final_state: str) -> dict[int, int]:
        permutations = {}
        for dest_pos in range(54):
            dest_char = final_state[dest_pos]
            src_pos = initial_state.index(dest_char)
            if dest_pos != src_pos:
                permutations[dest_pos] = src_pos
        return permutations

    def test_permutation_correctness(self) -> None:
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate('U')
        u_state = cube.state
        u_perms = self.get_permutations(initial_state, u_state)

        self.assertGreater(
            len(u_perms), 0,
            'U must produce permutations',
        )

        for dest, src in u_perms.items():
            self.assertIn(
                dest, range(54),
                f'Destination position { dest } invalid',
            )
            self.assertIn(
                src, range(54),
                f'Source position { src } invalid',
            )

    def analyze_move_permutations(self, move_name: str) -> dict[str, Any]:
        cube = self.create_numbered_cube()
        initial_state = cube.state

        cube.rotate(move_name)
        after_state = cube.state
        permutations = self.get_permutations(initial_state, after_state)

        return {
            'move': move_name,
            'permutation_count': len(permutations),
            'permutations': permutations,
            'cycles': self.find_cycles(permutations),
        }

    @staticmethod
    def find_cycles(permutations: dict[int, int]) -> list[list[int]]:
        visited = set()
        cycles = []

        for start in permutations:
            if start in visited:
                continue

            cycle = []
            current = start

            while current not in visited:
                visited.add(current)
                cycle.append(current)
                current = permutations.get(current, current)

                if current == start:
                    break

            if len(cycle) > 1:
                cycles.append(cycle)

        return cycles

    def test_cancellation_all_moves(self) -> None:
        moves = [
            'U', 'R', 'F', 'D', 'L', 'B',
            'x', 'y', 'z', 'M', 'E', 'S',
            'u', 'r', 'f', 'd', 'l', 'b',
        ]

        for move in moves:
            with self.subTest(move=move):
                cube = self.create_numbered_cube()
                initial_state = cube.state

                cube.rotate(move)
                cube.rotate(f"{ move }'")

                self.assertEqual(
                    cube.state, initial_state,
                    f"The move { move } does not cancel with { move }'",
                )

    def test_double_moves_cancellation(self) -> None:
        moves = [
            'U', 'R', 'F', 'D', 'L', 'B',
            'x', 'y', 'z', 'M', 'E', 'S',
            'u', 'r', 'f', 'd', 'l', 'b',
        ]

        for move in moves:
            with self.subTest(move=move):
                cube = self.create_numbered_cube()
                initial_state = cube.state

                cube.rotate(f'{ move }2')
                cube.rotate(f'{ move }2')

                self.assertEqual(
                    cube.state, initial_state,
                    f'The move { move }2 does not cancel with himself',
                )

    def test_all_basic_moves_produce_permutations(self) -> None:
        moves = ['U', 'R', 'F', 'D', 'L', 'B', 'M', 'E', 'S']

        for move in moves:
            with self.subTest(move=move):
                analysis = self.analyze_move_permutations(move)

                self.assertGreater(
                    analysis['permutation_count'], 0,
                    f'The move { move } does not produce permutation',
                )

                cycles = analysis['cycles']
                self.assertGreater(
                    len(cycles), 0,
                    f'The move { move } do not produce cycle',
                )

    def test_rotations_permutations(self) -> None:
        rotations = ['x', 'y', 'z']

        for rotation in rotations:
            with self.subTest(rotation=rotation):
                analysis = self.analyze_move_permutations(rotation)

                self.assertGreaterEqual(
                    analysis['permutation_count'], 24,
                    f'The rotation { rotation } affects too few positions',
                )

                cube = self.create_numbered_cube()
                initial_state = cube.state

                cube.rotate(rotation)
                cube.rotate(f"{rotation}'")

                self.assertEqual(
                    cube.state, initial_state,
                    f'The rotation { rotation } is not reversible',
                )

    def test_wide_moves_permutations(self) -> None:
        wide_moves = ['u', 'r', 'f', 'd', 'l', 'b']

        for move in wide_moves:
            with self.subTest(move=move):
                analysis = self.analyze_move_permutations(move)

                self.assertGreaterEqual(
                    analysis['permutation_count'], 16,
                    f'The wide move { move } affects too few positions',
                )

                cube = self.create_numbered_cube()
                initial_state = cube.state

                cube.rotate(move)
                cube.rotate(f"{ move }'")

                self.assertEqual(
                    cube.state, initial_state,
                    f'The move { move } is not reversible',
                )

    def test_all_move_variants(self) -> None:
        base_moves = [
            'U', 'R', 'F', 'D', 'L', 'B',
            'x', 'y', 'z', 'M', 'E', 'S',
            'u', 'r', 'f', 'd', 'l', 'b',
        ]

        for base_move in base_moves:
            for suffix in ['', "'", '2']:
                move = f'{ base_move }{ suffix }'
                with self.subTest(move=move):
                    cube = self.create_numbered_cube()

                    try:
                        cube.rotate(move)
                        after_state = cube.state

                        self.assertIsNotNone(
                            after_state,
                            f'The move { move } has failed',
                        )

                        if suffix != '2':
                            pass

                    except Exception as e:  # noqa: BLE001
                        self.fail(f'The move { move } has failed: { e }')

    def test_specific_move_permutations_u(self) -> None:
        analysis = self.analyze_move_permutations('U')

        self.assertEqual(
            analysis['permutation_count'], 20,
            'The move U must affect exactly 20 positions',
        )

        cycles = analysis['cycles']
        self.assertGreater(len(cycles), 0, 'U must produce cycles')

    def test_specific_move_permutations_m(self) -> None:
        analysis = self.analyze_move_permutations('M')

        self.assertEqual(
            analysis['permutation_count'], 12,
            'The move M must affect exactly 12 positions',
        )

        cycles = analysis['cycles']
        self.assertGreater(len(cycles), 0, 'M must produce cycles')

    def test_permutation_bijectivity(self) -> None:
        moves = [
            'U', 'R', 'F', 'D', 'L', 'B',
            'x', 'y', 'z', 'M', 'E', 'S',
        ]

        for move in moves:
            with self.subTest(move=move):
                analysis = self.analyze_move_permutations(move)
                permutations = analysis['permutations']

                sources = list(permutations.values())
                unique_sources = set(sources)

                self.assertEqual(
                    len(sources), len(unique_sources),
                    f'The move { move } has duplicated sources',
                )

                for dest, src in permutations.items():
                    self.assertIn(
                        dest, range(54),
                        f'Destination position { dest } invalid',
                    )
                    self.assertIn(
                        src, range(54),
                        f'Source position { src } invalid',
                    )



import unittest
from unittest.mock import patch

from cubing_algs.effects import EFFECTS
from cubing_algs.effects import FACE_POSITIONS
from cubing_algs.effects import brighten
from cubing_algs.effects import checkerboard
from cubing_algs.effects import chrome
from cubing_algs.effects import contrast
from cubing_algs.effects import copper
from cubing_algs.effects import diamond
from cubing_algs.effects import dim
from cubing_algs.effects import face_visible
from cubing_algs.effects import frosted
from cubing_algs.effects import get_position_factor
from cubing_algs.effects import global_light_position_factor
from cubing_algs.effects import glossy
from cubing_algs.effects import gold
from cubing_algs.effects import holographic
from cubing_algs.effects import load_effect
from cubing_algs.effects import load_single_effect
from cubing_algs.effects import matte
from cubing_algs.effects import neon
from cubing_algs.effects import noop
from cubing_algs.effects import parse_effect_name
from cubing_algs.effects import parse_effect_parameters
from cubing_algs.effects import plasma
from cubing_algs.effects import rainbow
from cubing_algs.effects import shine
from cubing_algs.effects import silver
from cubing_algs.effects import spiral
from cubing_algs.effects import stripes
from cubing_algs.effects import vintage


class TestPositioningFunctions(unittest.TestCase):
    """Test positioning and utility functions."""

    def test_face_positions_constant(self) -> None:
        """Test FACE_POSITIONS constant has correct structure."""
        self.assertEqual(len(FACE_POSITIONS), 6)
        for face_index in range(6):
            self.assertIn(face_index, FACE_POSITIONS)
            self.assertEqual(len(FACE_POSITIONS[face_index]), 2)
            self.assertIsInstance(FACE_POSITIONS[face_index][0], int)
            self.assertIsInstance(FACE_POSITIONS[face_index][1], int)

    def test_global_light_position_factor_basic(self) -> None:
        """Test global_light_position_factor with basic inputs."""
        result = global_light_position_factor(0, 3)
        self.assertIsInstance(result, float)
        self.assertGreaterEqual(result, 0.0)
        self.assertLessEqual(result, 1.0)

    def test_global_light_position_factor_different_cube_sizes(self) -> None:
        """Test global_light_position_factor with different cube sizes."""
        for cube_size in [2, 3, 4, 5]:
            result = global_light_position_factor(0, cube_size)
            self.assertIsInstance(result, float)
            self.assertGreaterEqual(result, 0.0)
            self.assertLessEqual(result, 1.0)

    def test_global_light_position_factor_all_faces(self) -> None:
        """Test global_light_position_factor across all faces."""
        cube_size = 3
        face_size = cube_size * cube_size

        for face_index in range(6):
            for local_index in range(face_size):
                facelet_index = face_index * face_size + local_index
                result = global_light_position_factor(facelet_index, cube_size)
                self.assertIsInstance(result, float)
                self.assertGreaterEqual(result, 0.0)
                self.assertLessEqual(result, 1.0)

    def test_global_light_position_factor_edge_cases(self) -> None:
        """Test global_light_position_factor with edge cases."""
        # Minimum cube size
        result = global_light_position_factor(0, 1)
        self.assertIsInstance(result, float)
        self.assertGreaterEqual(result, 0.0)
        self.assertLessEqual(result, 1.0)

        # Test with facelet indices at the edge of valid range
        cube_size = 3
        max_valid_index = 6 * cube_size * cube_size - 1  # 53 for 3x3 cube
        result = global_light_position_factor(max_valid_index, cube_size)
        self.assertIsInstance(result, float)
        self.assertGreaterEqual(result, 0.0)
        self.assertLessEqual(result, 1.0)

    def test_get_position_factor_default_parameters(self) -> None:
        """Test get_position_factor with default parameters."""
        result = get_position_factor(0, 3)
        self.assertIsInstance(result, float)
        self.assertGreaterEqual(result, 0.0)
        self.assertLessEqual(result, 1.0)

    def test_get_position_factor_facelet_modes(self) -> None:
        """Test get_position_factor with different facelet modes."""
        facelet_index = 4
        cube_size = 3

        # Local mode
        result_local = get_position_factor(
            facelet_index, cube_size, facelet_mode='local',
        )
        self.assertIsInstance(result_local, float)

        # Global mode
        result_global = get_position_factor(
            facelet_index, cube_size, facelet_mode='global',
        )
        self.assertIsInstance(result_global, float)

    def test_get_position_factor_position_modes(self) -> None:
        """Test get_position_factor with different position modes."""
        facelet_index = 4
        cube_size = 3

        # Numeral mode
        result_numeral = get_position_factor(
            facelet_index, cube_size, position_mode='numeral',
        )
        self.assertIsInstance(result_numeral, float)

        # Light mode
        result_light = get_position_factor(
            facelet_index, cube_size, position_mode='light',
        )
        self.assertIsInstance(result_light, float)

    def test_get_position_factor_mode_combinations(self) -> None:
        """Test get_position_factor with all mode combinations."""
        facelet_index = 4
        cube_size = 3

        modes = [
            {'facelet_mode': 'local', 'position_mode': 'numeral'},
            {'facelet_mode': 'global', 'position_mode': 'numeral'},
            {'facelet_mode': 'local', 'position_mode': 'light'},
            {'facelet_mode': 'global', 'position_mode': 'light'},
        ]

        for kwargs in modes:
            result = get_position_factor(facelet_index, cube_size, **kwargs)  # type: ignore[arg-type]
            self.assertIsInstance(result, float)
            self.assertGreaterEqual(result, 0.0)


class RGBTestCase(unittest.TestCase):

    def _validate_rgb_output(self, rgb: tuple[int, int, int]) -> None:
        """Helper to validate RGB output is properly clamped."""
        self.assertIsInstance(rgb, tuple)
        self.assertEqual(len(rgb), 3)
        for component in rgb:
            self.assertIsInstance(component, int)
            self.assertGreaterEqual(component, 0)
            self.assertLessEqual(component, 255)


class TestBasicEffects(RGBTestCase):
    """Test basic lighting and surface effects."""

    def setUp(self) -> None:
        """Set up common test data."""
        self.test_rgb = (128, 64, 192)
        self.facelet_index = 4
        self.cube_size = 3

    def test_shine_basic(self) -> None:
        """Test shine effect with basic parameters."""
        result = shine(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_shine_with_intensity(self) -> None:
        """Test shine effect with different intensity values."""
        intensities = [0.0, 0.3, 0.5, 0.8, 1.0]
        for intensity in intensities:
            result = shine(
                self.test_rgb,
                self.facelet_index,
                self.cube_size,
                intensity=intensity,
            )
            self._validate_rgb_output(result)

    def test_neon_basic(self) -> None:
        """Test neon effect with basic parameters."""
        result = neon(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_neon_with_parameters(self) -> None:
        """Test neon effect with different parameters."""
        result = neon(
            self.test_rgb,
            self.facelet_index,
            self.cube_size,
            intensity=0.8,
            saturation=1.5,
        )
        self._validate_rgb_output(result)

    def test_neon_black_input(self) -> None:
        """Test neon effect with black input (edge case)."""
        black_rgb = (0, 0, 0)
        result = neon(black_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_chrome_basic(self) -> None:
        """Test chrome effect with basic parameters."""
        result = chrome(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_chrome_with_parameters(self) -> None:
        """Test chrome effect with different parameters."""
        result = chrome(
            self.test_rgb,
            self.facelet_index,
            self.cube_size,
            intensity=0.7,
            metallic=0.8,
        )
        self._validate_rgb_output(result)

    def test_gold_basic(self) -> None:
        """Test gold effect with basic parameters."""
        result = gold(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_gold_with_warmth(self) -> None:
        """Test gold effect with different warmth values."""
        warmths = [0.0, 0.5, 1.0, 1.5]
        for warmth in warmths:
            result = gold(
                self.test_rgb,
                self.facelet_index,
                self.cube_size,
                warmth=warmth,
            )
            self._validate_rgb_output(result)

    def test_silver_basic(self) -> None:
        """Test silver effect with basic parameters."""
        result = silver(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_silver_with_intensity(self) -> None:
        """Test silver effect with different intensity values."""
        intensities = [0.1, 0.5, 0.9]
        for intensity in intensities:
            result = silver(
                self.test_rgb,
                self.facelet_index,
                self.cube_size,
                intensity=intensity,
            )
            self._validate_rgb_output(result)

    def test_copper_basic(self) -> None:
        """Test copper effect with basic parameters."""
        result = copper(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_copper_with_parameters(self) -> None:
        """Test copper effect with different parameters."""
        result = copper(
            self.test_rgb,
            self.facelet_index,
            self.cube_size,
            intensity=0.8,
            warmth=1.2,
        )
        self._validate_rgb_output(result)


class TestPatternEffects(RGBTestCase):
    """Test pattern-based effects."""

    def setUp(self) -> None:
        """Set up common test data."""
        self.test_rgb = (100, 150, 200)
        self.cube_size = 3

    def test_diamond_sparkle_positions(self) -> None:
        """Test diamond effect at sparkle positions."""
        # Test specific sparkle positions
        sparkle_positions = [(0, 0), (1, 1), (2, 2), (0, 2), (2, 0)]

        for row, col in sparkle_positions:
            facelet_index = row * self.cube_size + col
            result = diamond(self.test_rgb, facelet_index, self.cube_size)
            self._validate_rgb_output(result)
            # Sparkle positions should be brighter
            self.assertGreater(sum(result), sum(self.test_rgb))

    def test_diamond_non_sparkle_positions(self) -> None:
        """Test diamond effect at non-sparkle positions."""
        # Test non-sparkle position
        facelet_index = 1 * self.cube_size + 0  # Position (1, 0)
        result = diamond(self.test_rgb, facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_rainbow_basic(self) -> None:
        """Test rainbow effect with basic parameters."""
        for facelet_index in range(9):  # Test all positions on a 3x3 face
            result = rainbow(self.test_rgb, facelet_index, self.cube_size)
            self._validate_rgb_output(result)

    def test_checkerboard_pattern(self) -> None:
        """Test checkerboard effect creates alternating pattern."""
        results = []
        for facelet_index in range(9):  # Test all positions on a 3x3 face
            result = checkerboard(self.test_rgb, facelet_index, self.cube_size)
            self._validate_rgb_output(result)
            results.append(result)

        # Check that adjacent positions have different brightness
        # Position (0,0) and (0,1) should be different
        pos_00 = checkerboard(self.test_rgb, 0, self.cube_size)
        pos_01 = checkerboard(self.test_rgb, 1, self.cube_size)
        self.assertNotEqual(pos_00, pos_01)

    def test_checkerboard_with_intensity(self) -> None:
        """Test checkerboard effect with different intensities."""
        intensities = [0.2, 0.5, 0.8]
        for intensity in intensities:
            result = checkerboard(
                self.test_rgb, 0, self.cube_size, intensity=intensity,
            )
            self._validate_rgb_output(result)

    def test_stripes_directions(self) -> None:
        """Test stripes effect with different directions."""
        directions = ['horizontal', 'vertical', 'diagonal']
        for direction in directions:
            result = stripes(
                self.test_rgb, 4, self.cube_size, direction=direction,
            )
            self._validate_rgb_output(result)

    def test_stripes_with_parameters(self) -> None:
        """Test stripes effect with different parameters."""
        result = stripes(
            self.test_rgb,
            4,
            self.cube_size,
            direction='horizontal',
            frequency=3,
            intensity=0.6,
        )
        self._validate_rgb_output(result)

    def test_spiral_basic(self) -> None:
        """Test spiral effect with basic parameters."""
        for facelet_index in range(9):
            result = spiral(self.test_rgb, facelet_index, self.cube_size)
            self._validate_rgb_output(result)

    def test_spiral_with_intensity(self) -> None:
        """Test spiral effect with different intensities."""
        intensities = [0.1, 0.5, 0.9]
        for intensity in intensities:
            result = spiral(
                self.test_rgb, 4, self.cube_size, intensity=intensity,
            )
            self._validate_rgb_output(result)

    def test_plasma_basic(self) -> None:
        """Test plasma effect with basic parameters."""
        for facelet_index in range(9):
            result = plasma(self.test_rgb, facelet_index, self.cube_size)
            self._validate_rgb_output(result)

    def test_plasma_with_intensity(self) -> None:
        """Test plasma effect with different intensities."""
        intensities = [0.2, 0.4, 0.8]
        for intensity in intensities:
            result = plasma(
                self.test_rgb, 4, self.cube_size, intensity=intensity,
            )
            self._validate_rgb_output(result)


class TestSurfaceEffects(RGBTestCase):
    """Test surface finish effects."""

    def setUp(self) -> None:
        """Set up common test data."""
        self.test_rgb = (120, 80, 160)
        self.facelet_index = 4
        self.cube_size = 3

    def test_matte_basic(self) -> None:
        """Test matte effect with basic parameters."""
        result = matte(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)
        # Matte should reduce brightness
        self.assertLessEqual(sum(result), sum(self.test_rgb))

    def test_matte_ignores_position(self) -> None:
        """Test that matte effect ignores facelet position."""
        result1 = matte(self.test_rgb, 0, self.cube_size)
        result2 = matte(self.test_rgb, 8, self.cube_size)
        self.assertEqual(result1, result2)

    def test_matte_with_reduction(self) -> None:
        """Test matte effect with different reduction values."""
        reductions = [0.1, 0.3, 0.5, 0.8]
        for reduction in reductions:
            result = matte(
                self.test_rgb,
                self.facelet_index,
                self.cube_size,
                reduction=reduction,
            )
            self._validate_rgb_output(result)

    def test_glossy_basic(self) -> None:
        """Test glossy effect with basic parameters."""
        result = glossy(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_glossy_with_intensity(self) -> None:
        """Test glossy effect with different intensities."""
        intensities = [0.2, 0.5, 1.0]
        for intensity in intensities:
            result = glossy(
                self.test_rgb,
                self.facelet_index,
                self.cube_size,
                intensity=intensity,
            )
            self._validate_rgb_output(result)

    def test_frosted_basic(self) -> None:
        """Test frosted effect with basic parameters."""
        result = frosted(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_frosted_with_intensity(self) -> None:
        """Test frosted effect with different intensities."""
        intensities = [0.1, 0.4, 0.7]
        for intensity in intensities:
            result = frosted(
                self.test_rgb,
                self.facelet_index,
                self.cube_size,
                intensity=intensity,
            )
            self._validate_rgb_output(result)

    def test_holographic_basic(self) -> None:
        """Test holographic effect with basic parameters."""
        result = holographic(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_holographic_with_intensity(self) -> None:
        """Test holographic effect with different intensities."""
        intensities = [0.3, 0.6, 0.9]
        for intensity in intensities:
            result = holographic(
                self.test_rgb,
                self.facelet_index,
                self.cube_size,
                intensity=intensity,
            )
            self._validate_rgb_output(result)


class TestAdjustmentEffects(RGBTestCase):
    """Test color adjustment effects."""

    def setUp(self) -> None:
        """Set up common test data."""
        self.test_rgb = (128, 128, 128)
        self.facelet_index = 4
        self.cube_size = 3

    def test_dim_basic(self) -> None:
        """Test dim effect with basic parameters."""
        result = dim(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)
        # Dim should reduce brightness
        self.assertLessEqual(sum(result), sum(self.test_rgb))

    def test_dim_ignores_position(self) -> None:
        """Test that dim effect ignores facelet position."""
        result1 = dim(self.test_rgb, 0, self.cube_size)
        result2 = dim(self.test_rgb, 8, self.cube_size)
        self.assertEqual(result1, result2)

    def test_dim_with_factor(self) -> None:
        """Test dim effect with different factors."""
        factors = [0.1, 0.5, 0.7, 0.9]
        for factor in factors:
            result = dim(
                self.test_rgb,
                self.facelet_index,
                self.cube_size,
                factor=factor,
            )
            self._validate_rgb_output(result)

    def test_brighten_basic(self) -> None:
        """Test brighten effect with basic parameters."""
        result = brighten(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)
        # Brighten should increase brightness
        self.assertGreaterEqual(sum(result), sum(self.test_rgb))

    def test_brighten_ignores_position(self) -> None:
        """Test that brighten effect ignores facelet position."""
        result1 = brighten(self.test_rgb, 0, self.cube_size)
        result2 = brighten(self.test_rgb, 8, self.cube_size)
        self.assertEqual(result1, result2)

    def test_brighten_with_factor(self) -> None:
        """Test brighten effect with different factors."""
        factors = [1.1, 1.3, 1.5, 2.0]
        for factor in factors:
            result = brighten(
                self.test_rgb,
                self.facelet_index,
                self.cube_size,
                factor=factor,
            )
            self._validate_rgb_output(result)

    def test_contrast_basic(self) -> None:
        """Test contrast effect with basic parameters."""
        result = contrast(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_contrast_ignores_position(self) -> None:
        """Test that contrast effect ignores facelet position."""
        result1 = contrast(self.test_rgb, 0, self.cube_size)
        result2 = contrast(self.test_rgb, 8, self.cube_size)
        self.assertEqual(result1, result2)

    def test_contrast_with_factor(self) -> None:
        """Test contrast effect with different factors."""
        factors = [0.5, 1.0, 1.5, 2.0]
        for factor in factors:
            result = contrast(
                self.test_rgb,
                self.facelet_index,
                self.cube_size,
                factor=factor,
            )
            self._validate_rgb_output(result)

    def test_contrast_with_different_colors(self) -> None:
        """Test contrast effect with colors above and below middle gray."""
        # Test with bright color
        bright_rgb = (200, 200, 200)
        result_bright = contrast(bright_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result_bright)

        # Test with dark color
        dark_rgb = (50, 50, 50)
        result_dark = contrast(dark_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result_dark)

    def test_vintage_basic(self) -> None:
        """Test vintage effect with basic parameters."""
        result = vintage(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_vintage_ignores_position(self) -> None:
        """Test that vintage effect ignores facelet position."""
        result1 = vintage(self.test_rgb, 0, self.cube_size)
        result2 = vintage(self.test_rgb, 8, self.cube_size)
        self.assertEqual(result1, result2)

    def test_vintage_with_parameters(self) -> None:
        """Test vintage effect with different parameters."""
        result = vintage(
            self.test_rgb,
            self.facelet_index,
            self.cube_size,
            sepia=0.8,
            desaturation=0.5,
        )
        self._validate_rgb_output(result)

    def test_face_visible_basic(self) -> None:
        """Test face_visible effect with basic parameters."""
        result = face_visible(self.test_rgb, self.facelet_index, self.cube_size)
        self._validate_rgb_output(result)

    def test_face_visible_front_vs_back(self) -> None:
        """Test face_visible effect treats front and back faces differently."""
        # Front face (face index 0, 1, 2)
        front_facelet = 4  # Face 0
        result_front = face_visible(
            self.test_rgb, front_facelet, self.cube_size,
        )

        # Back face (face index 3, 4, 5)
        back_facelet = 3 * 9 + 4  # Face 3
        result_back = face_visible(self.test_rgb, back_facelet, self.cube_size)

        self._validate_rgb_output(result_front)
        self._validate_rgb_output(result_back)

        # Front faces should be brighter
        self.assertGreater(sum(result_front), sum(result_back))


class TestUtilityEffects(unittest.TestCase):
    """Test utility effects."""

    def setUp(self) -> None:
        """Set up common test data."""
        self.test_rgb = (100, 150, 200)
        self.facelet_index = 4
        self.cube_size = 3

    def test_noop_returns_unchanged(self) -> None:
        """Test that noop effect returns input unchanged."""
        result = noop(self.test_rgb, self.facelet_index, self.cube_size)
        self.assertEqual(result, self.test_rgb)

    def test_noop_ignores_all_parameters(self) -> None:
        """Test that noop effect ignores all parameters."""
        result1 = noop(self.test_rgb, 0, 2)
        result2 = noop(
            self.test_rgb, 100, 10, intensity=5.0, random_param='test',  # type: ignore[call-arg]
        )
        self.assertEqual(result1, self.test_rgb)
        self.assertEqual(result2, self.test_rgb)


class TestEdgeCases(RGBTestCase):
    """Test edge cases and boundary conditions."""

    def test_rgb_clamping_white(self) -> None:
        """Test RGB clamping with white input."""
        white_rgb = (255, 255, 255)
        effects_to_test = [shine, neon, chrome, gold, silver, copper, brighten]

        for effect in effects_to_test:
            result = effect(white_rgb, 0, 3)
            self._validate_rgb_output(result)

    def test_rgb_clamping_black(self) -> None:
        """Test RGB clamping with black input."""
        black_rgb = (0, 0, 0)
        effects_to_test = [shine, neon, chrome, gold, silver, copper, dim]

        for effect in effects_to_test:
            result = effect(black_rgb, 0, 3)
            self._validate_rgb_output(result)

    def test_different_cube_sizes(self) -> None:
        """Test effects with different cube sizes."""
        test_rgb = (128, 64, 192)
        cube_sizes = [1, 2, 3, 4, 5, 10]

        for cube_size in cube_sizes:
            for facelet_index in range(cube_size * cube_size):
                result = shine(test_rgb, facelet_index, cube_size)
                self._validate_rgb_output(result)

    def test_large_facelet_indices(self) -> None:
        """Test effects with large facelet indices within valid range."""
        test_rgb = (100, 100, 100)
        cube_size = 3
        # Test with indices from different faces but within valid range
        large_indices = [27, 45, 53]  # Face 3, 5, and last valid index

        for facelet_index in large_indices:
            result = diamond(test_rgb, facelet_index, cube_size)
            self._validate_rgb_output(result)

    def test_extreme_parameter_values(self) -> None:
        """Test effects with extreme parameter values."""
        test_rgb = (128, 128, 128)

        # Test with very high intensity
        result = shine(test_rgb, 0, 3, intensity=10.0)
        self._validate_rgb_output(result)

        # Test with zero intensity
        result = shine(test_rgb, 0, 3, intensity=0.0)
        self._validate_rgb_output(result)

        # Test with negative values
        result = dim(test_rgb, 0, 3, factor=-1.0)
        self._validate_rgb_output(result)

    def test_minimum_cube_size(self) -> None:
        """Test effects with minimum cube size of 1."""
        test_rgb = (128, 64, 192)
        cube_size = 1
        facelet_index = 0

        effects_to_test = [
            shine,
            neon,
            chrome,
            gold,
            silver,
            copper,
            diamond,
            rainbow,
            matte,
            glossy,
            frosted,
            checkerboard,
            stripes,
            spiral,
            plasma,
            holographic,
            dim,
            brighten,
            contrast,
            face_visible,
            vintage,
            noop,
        ]

        for effect in effects_to_test:
            result = effect(test_rgb, facelet_index, cube_size)
            self._validate_rgb_output(result)


class TestEffectsConfiguration(unittest.TestCase):
    """Test the EFFECTS configuration dictionary."""

    def test_effects_structure(self) -> None:
        """Test that EFFECTS dictionary has correct structure."""
        self.assertIsInstance(EFFECTS, dict)
        self.assertGreater(len(EFFECTS), 0)

        for effect_name, effect_config in EFFECTS.items():
            self.assertIsInstance(effect_name, str)
            self.assertIsInstance(effect_config, dict)
            self.assertIn('function', effect_config)
            self.assertTrue(callable(effect_config['function']))

    def test_effects_have_expected_functions(self) -> None:
        """Test that expected effects are present in EFFECTS."""
        expected_effects = [
            'shine',
            'soft',
            'gradient',
            'neon',
            'chrome',
            'gold',
            'silver',
            'copper',
            'diamond',
            'rainbow',
            'matte',
            'glossy',
            'frosted',
            'checkerboard',
            'h-stripes',
            'v-stripes',
            'd-stripes',
            'spiral',
            'plasma',
            'holographic',
            'dim',
            'brighten',
            'contrast',
            'vintage',
            'face-visible',
            'noop',
        ]

        for effect_name in expected_effects:
            self.assertIn(effect_name, EFFECTS)

    def test_effects_parameters_structure(self) -> None:
        """Test that effects parameters are properly structured."""
        for _effect_name, effect_config in EFFECTS.items():
            if 'parameters' in effect_config:
                params = effect_config['parameters']
                # Check that parameter values are reasonable types
                assert isinstance(params, dict)  # noqa: S101
                for param_name, param_value in params.items():
                    self.assertIsInstance(param_name, str)
                    self.assertIn(type(param_value), [int, float, str])

    def test_stripe_variants(self) -> None:
        """Test that stripe variants are properly configured."""
        stripe_variants = ['h-stripes', 'v-stripes', 'd-stripes']
        expected_directions = ['horizontal', 'vertical', 'diagonal']

        for variant, direction in zip(
            stripe_variants, expected_directions, strict=False,
        ):
            self.assertIn(variant, EFFECTS)
            self.assertEqual(EFFECTS[variant]['function'], stripes)
            params = EFFECTS[variant]['parameters']
            assert isinstance(params, dict)  # noqa: S101
            self.assertEqual(
                params['direction'], direction,
            )

    def test_shine_variants(self) -> None:
        """Test that shine variants are properly configured."""
        shine_variants = ['shine', 'soft', 'gradient']

        for variant in shine_variants:
            self.assertIn(variant, EFFECTS)
            self.assertEqual(EFFECTS[variant]['function'], shine)


class TestLoadEffect(unittest.TestCase):
    """Test the load_effect function."""

    def test_load_effect_valid_name(self) -> None:
        """Test load_effect with valid effect name."""
        effect_func = load_effect('shine', 'default')
        self.assertIsNotNone(effect_func)
        self.assertTrue(callable(effect_func))

    def test_load_effect_invalid_name(self) -> None:
        """Test load_effect with invalid effect name."""
        effect_func = load_effect('nonexistent', 'default')
        self.assertIsNone(effect_func)

    def test_load_effect_empty_name(self) -> None:
        """Test load_effect with empty effect name."""
        effect_func = load_effect('', 'default')
        self.assertIsNone(effect_func)

    def test_load_effect_none_name(self) -> None:
        """Test load_effect with None as effect name."""
        effect_func = load_effect(None, 'default')
        self.assertIsNone(effect_func)

    def test_load_effect_function_works(self) -> None:
        """Test that loaded effect function works correctly."""
        effect_func = load_effect('noop', 'default')
        self.assertIsNotNone(effect_func)
        assert effect_func is not None  # noqa: S101

        test_rgb = (100, 150, 200)
        result = effect_func(test_rgb, 0, 3)
        self.assertEqual(result, test_rgb)

    def test_load_effect_applies_parameters(self) -> None:
        """Test that loaded effect applies configured parameters."""
        effect_func = load_effect('dim', 'default')
        self.assertIsNotNone(effect_func)
        assert effect_func is not None  # noqa: S101

        test_rgb = (100, 100, 100)
        result = effect_func(test_rgb, 0, 3)

        # Should be dimmed according to default factor
        self.assertLess(sum(result), sum(test_rgb))

    def test_load_effect_palette_override(self) -> None:
        """Test load_effect with palette-specific parameters."""
        # This tests the palette override mechanism even though
        # no palette overrides exist in current EFFECTS
        effect_func = load_effect('shine', 'custom')
        self.assertIsNotNone(effect_func)
        assert effect_func is not None  # noqa: S101

        test_rgb = (100, 100, 100)
        result = effect_func(test_rgb, 0, 3)

        # Should still work even if palette doesn't exist
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 3)

    def test_load_effect_all_configured_effects(self) -> None:
        """Test that all configured effects can be loaded."""
        for effect_name in EFFECTS:
            effect_func = load_effect(effect_name, 'default')
            self.assertIsNotNone(
                effect_func, f'Failed to load effect: {effect_name}',
            )
            self.assertTrue(callable(effect_func))


class TestParameterParsing(unittest.TestCase):
    """Test parameter parsing functions."""

    def test_parse_effect_parameters_empty(self) -> None:
        """Test parsing empty parameter string."""
        result = parse_effect_parameters('')
        self.assertEqual(result, {})

    def test_parse_effect_parameters_single_int(self) -> None:
        """Test parsing single integer parameter."""
        result = parse_effect_parameters('intensity=5')
        self.assertEqual(result, {'intensity': 5})

    def test_parse_effect_parameters_single_float(self) -> None:
        """Test parsing single float parameter."""
        result = parse_effect_parameters('factor=0.75')
        self.assertEqual(result, {'factor': 0.75})

    def test_parse_effect_parameters_single_string(self) -> None:
        """Test parsing single string parameter."""
        result = parse_effect_parameters('direction=horizontal')
        self.assertEqual(result, {'direction': 'horizontal'})

    def test_parse_effect_parameters_quoted_string(self) -> None:
        """Test parsing quoted string parameter."""
        result = parse_effect_parameters('mode="light mode"')
        self.assertEqual(result, {'mode': 'light mode'})

        result = parse_effect_parameters("mode='dark mode'")
        self.assertEqual(result, {'mode': 'dark mode'})

    def test_parse_effect_parameters_boolean(self) -> None:
        """Test parsing boolean parameters."""
        result = parse_effect_parameters('enabled=true')
        self.assertEqual(result, {'enabled': True})

        result = parse_effect_parameters('disabled=false')
        self.assertEqual(result, {'disabled': False})

    def test_parse_effect_parameters_multiple(self) -> None:
        """Test parsing multiple parameters."""
        result = parse_effect_parameters(
            'intensity=0.8,factor=1.5,direction=vertical',
        )
        expected = {
            'intensity': 0.8, 'factor': 1.5,
            'direction': 'vertical',
        }
        self.assertEqual(result, expected)

    def test_parse_effect_parameters_mixed_types(self) -> None:
        """Test parsing mixed parameter types."""
        result = parse_effect_parameters(
            'intensity=0.8,frequency=3,enabled=true,mode=glossy',
        )
        expected = {
            'intensity': 0.8, 'frequency': 3,
            'enabled': True, 'mode': 'glossy',
        }
        self.assertEqual(result, expected)

    def test_parse_effect_parameters_negative_numbers(self) -> None:
        """Test parsing negative numbers."""
        result = parse_effect_parameters('offset=-5,factor=-0.3')
        expected = {'offset': -5, 'factor': -0.3}
        self.assertEqual(result, expected)

    def test_parse_effect_parameters_whitespace(self) -> None:
        """Test parsing with extra whitespace."""
        result = parse_effect_parameters('  intensity = 0.8 , factor = 1.5  ')
        expected = {'intensity': 0.8, 'factor': 1.5}
        self.assertEqual(result, expected)

    def test_parse_effect_parameters_invalid_format(self) -> None:
        """Test parsing with invalid parameter format."""
        result = parse_effect_parameters('intensity,factor=1.5')
        expected = {'factor': 1.5}  # Invalid param is ignored
        self.assertEqual(result, expected)

    def test_parse_effect_name_simple(self) -> None:
        """Test parsing simple effect name without parameters."""
        name, params = parse_effect_name('shine')
        self.assertEqual(name, 'shine')
        self.assertEqual(params, {})

    def test_parse_effect_name_with_parameters(self) -> None:
        """Test parsing effect name with parameters."""
        name, params = parse_effect_name('shine(intensity=0.8)')
        self.assertEqual(name, 'shine')
        self.assertEqual(params, {'intensity': 0.8})

    def test_parse_effect_name_with_multiple_parameters(self) -> None:
        """Test parsing effect name with multiple parameters."""
        name, params = parse_effect_name('chrome(intensity=0.7,metallic=0.9)')
        self.assertEqual(name, 'chrome')
        self.assertEqual(params, {'intensity': 0.7, 'metallic': 0.9})

    def test_parse_effect_name_empty_parameters(self) -> None:
        """Test parsing effect name with empty parameter list."""
        name, params = parse_effect_name('neon()')
        self.assertEqual(name, 'neon')
        self.assertEqual(params, {})

    def test_parse_effect_name_whitespace(self) -> None:
        """Test parsing effect name with whitespace."""
        name, params = parse_effect_name('  gold( intensity=0.6 ) ')
        self.assertEqual(name, 'gold')
        self.assertEqual(params, {'intensity': 0.6})

    def test_parse_effect_name_invalid_syntax(self) -> None:
        """Test parsing effect name with invalid syntax triggers fallback."""
        # Test case with unmatched parentheses that won't match the regex
        name, params = parse_effect_name('shine(intensity=0.8')
        self.assertEqual(name, 'shine(intensity=0.8')  # Returns as-is
        self.assertEqual(params, {})

        # Test case with multiple unmatched parentheses
        name, params = parse_effect_name('effect((invalid')
        self.assertEqual(name, 'effect((invalid')
        self.assertEqual(params, {})


class TestEnhancedLoadEffect(unittest.TestCase):
    """Test enhanced load_effect function with chaining and parameters."""

    def test_load_effect_single_effect(self) -> None:
        """Test loading single effect without changes."""
        effect_func = load_effect('noop', 'default')
        self.assertIsNotNone(effect_func)
        assert effect_func is not None  # noqa: S101

        test_rgb = (100, 150, 200)
        result = effect_func(test_rgb, 0, 3)
        self.assertEqual(result, test_rgb)

    def test_load_effect_single_with_custom_params(self) -> None:
        """Test loading single effect with custom parameters."""
        effect_func = load_effect('dim(factor=0.5)', 'default')
        self.assertIsNotNone(effect_func)
        assert effect_func is not None  # noqa: S101

        test_rgb = (100, 100, 100)
        result = effect_func(test_rgb, 0, 3)

        # Should be dimmed by factor 0.5
        expected = (50, 50, 50)
        self.assertEqual(result, expected)

    def test_load_effect_chained_effects(self) -> None:
        """Test loading chained effects."""
        effect_func = load_effect('brighten|dim', 'default')
        self.assertIsNotNone(effect_func)
        assert effect_func is not None  # noqa: S101

        test_rgb = (100, 100, 100)
        result = effect_func(test_rgb, 0, 3)

        # Should apply brighten then dim
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 3)

    def test_load_effect_chained_with_params(self) -> None:
        """Test loading chained effects with custom parameters."""
        effect_func = load_effect(
            'brighten(factor=2.0)|dim(factor=0.5)',
            'default',
        )
        self.assertIsNotNone(effect_func)
        assert effect_func is not None  # noqa: S101

        test_rgb = (100, 100, 100)
        result = effect_func(test_rgb, 0, 3)

        # Should apply brighten factor 2.0 then dim factor 0.5
        # (100 * 2.0) * 0.5 = 100, so should return to original
        expected = (100, 100, 100)
        self.assertEqual(result, expected)

    def test_load_effect_multiple_chained(self) -> None:
        """Test loading multiple chained effects."""
        effect_func = load_effect('noop|noop|noop', 'default')
        self.assertIsNotNone(effect_func)
        assert effect_func is not None  # noqa: S101

        test_rgb = (100, 150, 200)
        result = effect_func(test_rgb, 0, 3)
        # Chain of noops should return original
        self.assertEqual(result, test_rgb)

    def test_load_effect_invalid_in_chain(self) -> None:
        """Test loading chain with invalid effect."""
        effect_func = load_effect('shine|invalid_effect|dim', 'default')
        # Should still work with valid effects
        self.assertIsNotNone(effect_func)
        assert effect_func is not None  # noqa: S101

        test_rgb = (100, 100, 100)
        result = effect_func(test_rgb, 0, 3)
        # Should apply shine and dim, skipping invalid
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 3)

    def test_load_effect_all_invalid_chain(self) -> None:
        """Test loading chain with all invalid effects."""
        effect_func = load_effect('invalid1|invalid2', 'default')
        self.assertIsNone(effect_func)

    def test_load_effect_empty_chain(self) -> None:
        """Test loading empty effect chain."""
        effect_func = load_effect('', 'default')
        self.assertIsNone(effect_func)

    def test_load_effect_complex_params(self) -> None:
        """Test loading effect with complex parameter combinations."""
        effect_func = load_effect(
            'd-stripes(frequency=3,intensity=0.6)',
            'default',
        )
        self.assertIsNotNone(effect_func)
        assert effect_func is not None  # noqa: S101

        test_rgb = (100, 100, 100)
        result = effect_func(test_rgb, 0, 3)
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 3)

    def test_load_effect_parameter_override(self) -> None:
        """Test that custom parameters override default ones."""
        # Load with very low intensity
        effect_func1 = load_effect('dim(factor=0.1)', 'default')
        assert effect_func1 is not None  # noqa: S101

        # Load with very high intensity
        effect_func2 = load_effect('dim(factor=0.9)', 'default')
        assert effect_func2 is not None  # noqa: S101

        test_rgb = (100, 100, 100)
        result1 = effect_func1(test_rgb, 0, 3)
        result2 = effect_func2(test_rgb, 0, 3)

        # Results should be different due to different factors
        self.assertNotEqual(result1, result2)

    def test_load_effect_whitespace_handling(self) -> None:
        """Test that whitespace in effect names is handled correctly."""
        effect_func = load_effect('  shine  |  dim  ', 'default')
        self.assertIsNotNone(effect_func)

    def test_load_single_effect_palette_specific_params(self) -> None:
        """Test load_single_effect with palette-specific parameters."""
        # Create a mock effect config with palette-specific parameters
        mock_effect_config = {
            'function': noop,
            'parameters': {'base_param': 1.0},
            'default': {'palette_param': 2.0},  # Palette-specific parameter
        }

        with patch.dict(
                'cubing_algs.effects.EFFECTS',
                {'test_effect': mock_effect_config},
        ):
            effect_func = load_single_effect('test_effect', {}, 'default')
            self.assertIsNotNone(effect_func)
            assert effect_func is not None  # noqa: S101

            # Test that the effect function works
            result = effect_func((100, 100, 100), 0, 3)
            self.assertEqual(result, (100, 100, 100))  # noop returns unchanged

        test_rgb = (100, 100, 100)
        result = effect_func(test_rgb, 0, 3)
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 3)

    def test_load_single_effect_palette_override_non_dict(self) -> None:
        """Test load_single_effect with non-dict palette override."""
        # Create a mock effect config with a non-dict palette override
        mock_effect_config = {
            'function': noop,
            'parameters': {'base_param': 1.0},
            'default': 'not_a_dict',  # Non-dict value that should be ignored
        }

        with patch.dict(
                'cubing_algs.effects.EFFECTS',
                {'test_effect': mock_effect_config},
        ):
            effect_func = load_single_effect('test_effect', {}, 'default')
            self.assertIsNotNone(effect_func)
            assert effect_func is not None  # noqa: S101

            # Test that the effect function works
            result = effect_func((100, 100, 100), 0, 3)
            self.assertEqual(result, (100, 100, 100))  # noop returns unchanged

    def test_load_effect_real_world_combinations(self) -> None:
        """Test real-world effect combinations."""
        combinations = [
            'shine(intensity=0.8)|chrome(metallic=0.7)',
            'gold(warmth=1.2)|glossy(intensity=0.9)',
            'checkerboard(intensity=0.3)|vintage(sepia=0.4)',
            'h-stripes(frequency=2)|rainbow',
            'dim(factor=0.8)|contrast(factor=1.5)|brighten(factor=1.1)',
        ]

        test_rgb = (128, 128, 128)
        for combination in combinations:
            effect_func = load_effect(combination, 'default')
            self.assertIsNotNone(effect_func, f'Failed to load: {combination}')
            assert effect_func is not None  # noqa: S101

            result = effect_func(test_rgb, 4, 3)
            self.assertIsInstance(result, tuple)
            self.assertEqual(len(result), 3)
            for component in result:
                self.assertGreaterEqual(component, 0)
                self.assertLessEqual(component, 255)



from cubing_algs.algorithm import Algorithm
from cubing_algs.move import Move


def mirror_moves(old_moves: Algorithm) -> Algorithm:
    """
    Create the mirror inverse of an algorithm.

    Reverses the order of moves and inverts each move to create
    the sequence that undoes the original algorithm.
    """
    moves: list[Move] = []

    for move in reversed(old_moves):
        moves.append(move.inverted)

    return Algorithm(moves)
